from running.benchmark import JavaBenchmark
from running.modifier import *
from running.config import Configuration
from running.runtime import OpenJDK


def test_jvm_arg():
    j = JVMArg(name="j", val='-Xms100M -D"foo bar"')
    assert j.val == ["-Xms100M", "-Dfoo bar"]


def test_jvm_classpath():
    j = JVMClasspath(name="j", val='/bin /foo "/Users/John Citizen/"')
    assert j.val == ["/bin", "/foo", "/Users/John Citizen/"]

    jb = JavaBenchmark(
        jvm_args=[], program_args=[], cp=["fizzbuzz"], suite_name="dacapo", name="fop"
    )
    jb = jb.attach_modifiers([j])
    assert jb.cp == ["fizzbuzz", "/bin", "/foo", "/Users/John Citizen/"]


def test_jvm_classpath_append():
    j = JVMClasspathAppend(name="j", val='/bin /foo "/Users/John Citizen/"')
    assert j.val == ["/bin", "/foo", "/Users/John Citizen/"]

    jb = JavaBenchmark(
        jvm_args=[], program_args=[], cp=["fizzbuzz"], suite_name="dacapo", name="fop"
    )
    jb = jb.attach_modifiers([j])
    assert jb.cp == ["fizzbuzz", "/bin", "/foo", "/Users/John Citizen/"]


def test_jvm_classpath_prepend():
    j = JVMClasspathPrepend(name="j", val='/bin /foo "/Users/John Citizen/"')
    assert j.val == ["/bin", "/foo", "/Users/John Citizen/"]

    jb = JavaBenchmark(
        jvm_args=[], program_args=[], cp=["fizzbuzz"], suite_name="dacapo", name="fop"
    )
    jb = jb.attach_modifiers([j])
    assert jb.cp == ["/bin", "/foo", "/Users/John Citizen/", "fizzbuzz"]


def test_program_arg():
    p = ProgramArg(name="p", val='/bin /foo "/Users/John Citizen/"')
    assert p.val == ["/bin", "/foo", "/Users/John Citizen/"]


def test_expand_value_opts():
    p = EnvVar(name="path", var="PATH", val="{0}:{1}")
    assert p.val == "{0}:{1}"
    p = p.apply_value_opts(value_opts=["/bin", "/sbin"])
    assert p.val == "/bin:/sbin"


def test_modifier_set():
    c = Configuration(
        {
            "modifiers": {
                "a": {"type": "JVMArg", "val": "-XX:GC={0}"},
                "b": {"type": "EnvVar", "var": "FOO", "val": "BAR"},
                "c": {"type": "EnvVar", "var": "FIZZ", "val": "BUZZ"},
                "set": {"type": "ModifierSet", "val": "a-{0}|b"},
                "set_nested": {"type": "ModifierSet", "val": "set-{0}|c"},
            }
        }
    )
    c.resolve_class()
    mods = c.get("modifiers")["set"].apply_value_opts(value_opts=["NoGC"]).flatten(c)
    mods = (
        c.get("modifiers")["set_nested"]
        .apply_value_opts(value_opts=["NoGC"])
        .flatten(c)
    )
    assert len(mods) == 3


def test_no_includes_no_excludes():
    c = Configuration(
        {"modifiers": {"a": {"type": "EnvVar", "var": "FOO", "val": "BAR"}}}
    )
    c.resolve_class()
    m = c.get("modifiers")["a"]
    assert m.should_attach("dacapo2006", "fop")


def test_excludes():
    c = Configuration(
        {
            "modifiers": {
                "a": {
                    "type": "EnvVar",
                    "var": "FOO",
                    "val": "BAR",
                    "excludes": {"dacapo2006": ["fop"]},
                }
            }
        }
    )
    c.resolve_class()
    m = c.get("modifiers")["a"]
    assert not m.should_attach("dacapo2006", "fop")
    assert m.should_attach("dacapo2006", "avrora")


def test_includes():
    c = Configuration(
        {
            "modifiers": {
                "a": {
                    "type": "EnvVar",
                    "var": "FOO",
                    "val": "BAR",
                    "includes": {"dacapo2006": ["fop"]},
                }
            }
        }
    )
    c.resolve_class()
    m = c.get("modifiers")["a"]
    assert m.should_attach("dacapo2006", "fop")
    assert not m.should_attach("dacapo2006", "avrora")


def test_includes_excludes():
    c = Configuration(
        {
            "modifiers": {
                "a": {
                    "type": "EnvVar",
                    "var": "FOO",
                    "val": "BAR",
                    "includes": {"dacapo2006": ["fop", "hsqldb"]},
                    "excludes": {"dacapo2006": ["fop"]},
                }
            }
        }
    )
    c.resolve_class()
    m = c.get("modifiers")["a"]
    assert not m.should_attach("dacapo2006", "fop")
    assert not m.should_attach("dacapo2006", "avrora")
    assert m.should_attach("dacapo2006", "hsqldb")


def test_envvar():
    jb = JavaBenchmark(
        jvm_args=[], program_args=[], cp=[], suite_name="dacapo", name="fop"
    )
    jb = jb.attach_modifiers([EnvVar(name="foobar", var="FOOBAR", val="$HOME")])
    jb = jb.attach_modifiers(
        [
            EnvVar(
                name="fizzbuzz",
                var="FIZZBUZZ",
                val="$DAHKDLHDIWHEIUWHEIWEHIJHDJKAGDKJADGUQDGIQUWDGI",
            )
        ]
    )
    openjdk = OpenJDK(
        release=11, home="/usr/lib/jvm/temurin-11-jdk-amd64", name="temurin-11"
    )
    assert "$HOME" not in jb.to_string(openjdk)
    assert "$DAHKDLHDIWHEIUWHEIWEHIJHDJKAGDKJADGUQDGIQUWDGI" in jb.to_string(openjdk)



#!/usr/env/bin python3
import logging
import argparse

from running.__version__ import __VERSION__
from running.command import fillin, runbms, minheap, log_preprocessor
from running.suite import set_dry_run
import importlib.resources
import os

logger = logging.getLogger(__name__)

MODULES = [fillin, runbms, minheap, log_preprocessor]


def setup_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="change logging level to DEBUG"
    )
    parser.add_argument(
        "--version", action="version", version="running {}".format(__VERSION__)
    )
    parser.add_argument("-d", "--dry-run", action="store_true", help="dry run")
    subparsers = parser.add_subparsers()
    for m in MODULES:
        m.setup_parser(subparsers)
    return parser


def main():
    parsers = setup_parser()
    args = vars(parsers.parse_args())

    # Config root logger
    if args.get("verbose") == True:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO
    logging.basicConfig(
        format="[%(levelname)s] %(asctime)s %(filename)s:%(lineno)d %(message)s",
        level=log_level,
    )

    if args.get("dry_run") == True:
        set_dry_run(True)
    with importlib.resources.path(__package__, "config") as config_path:
        os.environ["RUNNING_NG_PACKAGE_DATA"] = str(config_path)
        for m in MODULES:
            if m.run(args):
                break
        else:
            parsers.print_help()


if __name__ == "__main__":
    main()



from running.modifier import JVMArg, Modifier, JSArg, EnvVar
from typing import Any, Dict, List, Union
from pathlib import Path
import logging
from running.util import register
import os.path


class Runtime(object):
    CLS_MAPPING: Dict[str, Any]
    CLS_MAPPING = {}

    def __init__(self, name: str, **kwargs):
        self.name = name

    @staticmethod
    def from_config(name: str, config: Dict[str, str]) -> Any:
        return Runtime.CLS_MAPPING[config["type"]](name=name, **config)

    def get_executable(self) -> Union[str, Path]:
        raise NotImplementedError

    def get_heapsize_modifiers(self, size: int) -> List[Modifier]:
        raise NotImplementedError

    def is_oom(self, _output: bytes) -> bool:
        raise NotImplementedError


class DummyRuntime(Runtime):
    def __init__(self, executable: str):
        super().__init__(name="dummy")
        self.executable = executable

    def get_executable(self) -> Union[str, Path]:
        return self.executable

    def is_oom(self, _output: bytes) -> bool:
        return False


@register(Runtime)
class NativeExecutable(Runtime):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_executable(self) -> Union[str, Path]:
        return ""

    def is_oom(self, _output: bytes) -> bool:
        return False


class JVM(Runtime):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_executable(self) -> Path:
        raise NotImplementedError

    def __str__(self):
        return "JVM {}".format(self.name)

    def get_heapsize_modifiers(self, size: int) -> List[Modifier]:
        size_str = "{}M".format(size)
        heapsize = JVMArg(
            name="heap{}".format(size_str),
            val="-Xms{} -Xmx{}".format(size_str, size_str),
        )
        return [heapsize]

    def is_oom(self, output: bytes) -> bool:
        for pattern in [
            b"Allocation Failed",
            b"OutOfMemoryError",
            b"ran out of memory",
            b"panicked at 'Out of memory!'",
        ]:
            if pattern in output:
                return True
        return False


@register(Runtime)
class OpenJDK(JVM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.release = kwargs["release"]
        try:
            self.release = int(self.release)
        except ValueError:
            raise TypeError("The release of an OpenJDK has to be int-like")
        self.home: Path
        self.home = Path(os.path.expandvars(kwargs["home"]))
        if not self.home.exists():
            logging.warning("OpenJDK home {} doesn't exist".format(self.home))
        self.executable = self.home / "bin" / "java"
        if not self.executable.exists():
            logging.warning("{} not found in OpenJDK home".format(self.executable))
        self.executable = self.executable.absolute()

    def get_executable(self) -> Path:
        return self.executable

    def __str__(self):
        return "{} OpenJDK {} {}".format(super().__str__(), self.release, self.home)


@register(Runtime)
class JikesRVM(JVM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.home: Path
        self.home = Path(os.path.expandvars(kwargs["home"]))
        if not self.home.exists():
            logging.warning("JikesRVM home {} doesn't exist".format(self.home))
        self.executable = self.home / "rvm"
        if not self.home.exists():
            logging.warning("{} not found in JikesRVM home".format(self.executable))
        self.executable = self.executable.absolute()

    def get_executable(self) -> Path:
        return self.executable

    def __str__(self):
        return "{} JikesRVM {}".format(super().__str__(), self.home)


class JavaScriptRuntime(Runtime):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.executable: Path
        self.executable = Path(os.path.expandvars(kwargs["executable"]))
        if not self.executable.exists():
            logging.warning(
                "JavaScriptRuntime executable {} doesn't exist".format(self.executable)
            )
        self.executable = self.executable.absolute()

    def get_executable(self) -> Path:
        return self.executable


@register(Runtime)
class D8(JavaScriptRuntime):
    def __str__(self):
        return "{} d8 {}".format(super().__str__(), self.executable)

    def get_heapsize_modifiers(self, size: int) -> List[Modifier]:
        size_str = "{}".format(size)
        heapsize = JSArg(
            name="heap{}".format(size_str),
            val="--initial-heap-size={} --max-heap-size={}".format(size_str, size_str),
        )
        return [heapsize]

    def is_oom(self, output: bytes) -> bool:
        # The format is "Fatal javascript OOM in ..." or "Fatal JavaScript out of memory"
        # such as "Fatal javascript OOM in Reached heap limit"
        # or "Fatal javascript OOM in Ineffective mark-compacts near heap limit"
        # or "Fatal JavaScript out of memory: Reached heap limit"
        for pattern in [b"Fatal javascript OOM in", b"Fatal JavaScript out of memory"]:
            if pattern in output:
                return True
        return False


@register(Runtime)
class SpiderMonkey(JavaScriptRuntime):
    def __str__(self):
        return "{} SpiderMonkey {}".format(super().__str__(), self.executable)

    def get_heapsize_modifiers(self, size: int) -> List[Modifier]:
        size_str = "{}".format(size)
        # FIXME doesn't seem to be working
        heapsize = JSArg(
            name="heap{}".format(size_str), val="--available-memory={}".format(size_str)
        )
        return [heapsize]

    def is_oom(self, output: bytes) -> bool:
        # FIXME not sure how to check for OOM for SpiderMonkey yet
        return False


@register(Runtime)
class JavaScriptCore(JavaScriptRuntime):
    def __str__(self):
        return "{} JavaScriptCore {}".format(super().__str__(), self.executable)

    def get_heapsize_modifiers(self, size: int) -> List[Modifier]:
        size_str = "{}".format(size)
        # FIXME doesn't seem to be working
        heapsize = JSArg(
            name="heap{}".format(size_str), val="--gcMaxHeapSize={}".format(size_str)
        )
        return [heapsize]

    def is_oom(self, output: bytes) -> bool:
        # FIXME not sure how to check for OOM for JavaScriptCore yet
        return False


class Julia(Runtime):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.executable: Path
        self.executable = Path(os.path.expandvars(kwargs["executable"]))
        if not self.executable.exists():
            logging.warning("Julia executable {} doesn't exist".format(self.executable))
        self.executable = self.executable.absolute()

    def get_executable(self) -> Path:
        return self.executable

    def __str__(self):
        return "Julia {} {}".format(self.name, self.executable)


@register(Runtime)
class JuliaMMTK(Julia):
    def get_heapsize_modifiers(self, size: int) -> List[Modifier]:
        # size in MB
        size_str = "{}".format(size)
        min = EnvVar(
            name="minheap{}".format(size_str), var="MMTK_MIN_HSIZE", val=size_str
        )
        max = EnvVar(
            name="maxheap{}".format(size_str), var="MMTK_MAX_HSIZE", val=size_str
        )
        return [min, max]

    def __str__(self):
        return "{} with MMTk".format(super().__str__())

    def is_oom(self, output: bytes) -> bool:
        return b"Out of Memory!" in output


@register(Runtime)
class JuliaStock(Julia):
    def get_heapsize_modifiers(self, size: int) -> List[Modifier]:
        return []

    def __str__(self):
        return "{} stock version".format(super().__str__())

    def is_oom(self, output: bytes) -> bool:
        return False



#!/usr/bin/env python3
import sys
import os
import shutil
import gzip
import re
import glob
import subprocess
import itertools

ADVICE_EXTS = ["ca", "dc", "ec"]
advice_folder = sys.argv[1]


JikesRVM_HEADER = (
    "============================ MMTk Statistics Totals ============================"
)
JikesRVM_FOOTER = (
    "------------------------------ End MMTk Statistics -----------------------------"
)


def extract_blocks(lines, header, footer):
    lines = [l.decode("ascii") for l in lines]
    i = 0
    blocks = []
    while i < len(lines):
        if header in lines[i]:
            i += 1
            block = []
            while footer not in lines[i]:
                block.append(lines[i])
                i += 1
            blocks.append(block)
        i += 1
    return blocks


def cleanse(filename):
    sed_pattern = "s/{urls[^}]*}//g"
    if sys.platform == "linux" or sys.platform == "linux2":
        cmd = ["sed", "-i", sed_pattern, filename]
    elif sys.platform == "darwin":
        cmd = ["sed", "-i", "", sed_pattern, filename]
    print("run({})".format(" ".join(cmd)))
    subprocess.run(cmd)


def select_best_invocation(scenario):
    filename = "{}.log.gz".format(scenario)
    metrics = []
    with gzip.open(os.path.join(advice_folder, filename)) as log_file:
        stats_blocks = extract_blocks(log_file, JikesRVM_HEADER, JikesRVM_FOOTER)
        for stats_block in stats_blocks:
            stats = dict(zip(stats_block[0].split("\t"), stats_block[1].split("\t")))
            metrics.append(float(stats["time.gc"]) + float(stats["time.mu"]))
    if not metrics:
        print("No metric is found")
        return -1
    print("Metrics for {}: {}".format(scenario, metrics))
    _, idx = min([(val, idx) for (idx, val) in enumerate(metrics)])
    return idx


def select_advice_file(scenario, best_invocation):
    if not best_invocation >= 0:
        return
    benchmark_name = scenario.split(".")[0]
    for ext in ADVICE_EXTS:
        src = "{}.{}.{}".format(scenario, best_invocation, ext)
        src = os.path.join(advice_folder, src)
        dst = "{}.{}".format(benchmark_name, ext)
        dst = os.path.join(advice_folder, dst)
        print("Copying {} to {}".format(src, dst))
        shutil.copyfile(src, dst)
        cleanse(dst)


def main():
    scenario_logs = glob.glob(os.path.join(advice_folder, "*.log.gz"))
    scenarios = [os.path.basename(s).replace(".log.gz", "") for s in scenario_logs]
    print("Found scenarios {}".format(scenarios))
    for scenario in scenarios:
        print("Processing scenario {}".format(scenario))
        best_invocation = select_best_invocation(scenario)
        print("Best invocation for scenario {} is {}".format(scenario, best_invocation))
        select_advice_file(scenario, best_invocation)


if __name__ == "__main__":
    main()



import logging
from typing import (
    DefaultDict,
    Dict,
    List,
    Any,
    Optional,
    Set,
    Tuple,
    BinaryIO,
    TYPE_CHECKING,
)
from running.suite import BenchmarkSuite, is_dry_run
from running.benchmark import Benchmark, SubprocessrExit
from running.config import Configuration
from pathlib import Path
from running.util import (
    parse_config_str,
    system,
    get_logged_in_users,
    config_index_to_chr,
    config_str_encode,
)
import socket
from datetime import datetime
from running.runtime import Runtime
import tempfile
import subprocess
import os
from running.command.fillin import fillin
import math
import yaml

if TYPE_CHECKING:
    from running.plugin.runbms import RunbmsPlugin
from running.__version__ import __VERSION__

configuration: Configuration
minheap_multiplier: float
remote_host: Optional[str]
skip_oom: Optional[int]
skip_timeout: Optional[int]
skip_log_compression: bool = False
plugins: Dict[str, Any]
resume: Optional[str]


def setup_parser(subparsers):
    f = subparsers.add_parser("runbms")
    f.set_defaults(which="runbms")
    f.add_argument("LOG_DIR", type=Path)
    f.add_argument("CONFIG", type=Path)
    f.add_argument("N", type=int, nargs="?")
    f.add_argument("n", type=int, nargs="*")
    f.add_argument("-i", "--invocations", type=int)
    f.add_argument("-s", "--slice", type=str)
    f.add_argument("-p", "--id-prefix")
    f.add_argument("-m", "--minheap-multiplier", type=float)
    f.add_argument("--skip-oom", type=int)
    f.add_argument("--skip-timeout", type=int)
    f.add_argument("--resume", type=str)
    f.add_argument("--workdir", type=Path)
    f.add_argument(
        "--skip-log-compression", action="store_true", help="Skip compressing log files"
    )


def getid() -> str:
    host = socket.gethostname()
    now = datetime.now()
    timestamp = now.strftime("%Y-%m-%d-%a-%H%M%S")
    return "{}-{}".format(host, timestamp)


def spread(spread_factor: int, N: int, n: int) -> float:
    """Spread the numbers

    For example, when we have N = 8, n = 0, 1, ..., 7, N,
    we can have fractions of form n / N with equal distances between them:
    0/8, 1/8, 2/8, ..., 7/8, 8/8

    Sometimes it's desirable to transform such sequence with finer steppings
    at the start than at the end.
    The idea is that when the values are small, even small changes can greatly
    affect the behaviour of a system (such as the heap sizes of GC), so we want
    to explore smaller values at a finer granularity.

    For each n, we define
    n' = n + spread_factor / (N - 1) * (1 + 2 + ... + n-2 + n-1)
    For a consecutive pair (n-1, n), we can see that they are additionally
    (n-1) / (N-1) * spread_factor apart.
    For a special case when n = N, we have N' = N + N/2 * spread_factor,
    and (N'-1, N') is spread_factor apart.

    Let's concretely use spread_factor = 1 as as example.
    Let N and n be the same.
    So we have spread_factor / (N - 1) = 1/7
    n = 0, n' = 0,
    n = 1, n' = 1.
    n = 2, n' = 2 + 1/7
    n = 3, n' = 3 + 1/7 + 2/7 = 3 + 3/7
    ...
    n = 7, n' = 7 + 1/7 + 2/7 + ... + 6/7 = 7 + 21/7 = 10
    n = 8, n' = 8 + 1/7 + 2/7 + ... + 7/7 = 8 + 28/7 = 12

    Parameters
    ----------
    N : int
        Denominator
    spread_factor : int
        How much coarser is the spacing at the end relative to start
        spread_factor = 0 doesn't change the sequence at all
    n: int
        Nominator
    """
    sum_1_n_minus_1 = (n * n - n) / 2
    return n + spread_factor / (N - 1) * sum_1_n_minus_1


def hfac_str(hfac: float) -> str:
    return str(int(hfac * 1000))


def get_heapsize(hfac: float, minheap: int) -> int:
    return round(minheap * hfac * minheap_multiplier)


def get_hfacs(
    heap_range: int, spread_factor: int, N: int, ns: List[int]
) -> List[float]:
    start = 1.0
    end = float(heap_range)
    divisor = spread(spread_factor, N, N) / (end - start)
    return [spread(spread_factor, N, n) / divisor + start for n in ns]


def run_benchmark_with_config(
    c: str, b: Benchmark, runbms_dir: Path, size: Optional[int], fd: Optional[BinaryIO]
) -> Tuple[bytes, SubprocessrExit]:
    runtime, mods = parse_config_str(configuration, c)
    mod_b = b.attach_modifiers(mods)
    mod_b = mod_b.attach_modifiers(b.get_runtime_specific_modifiers(runtime))
    if size is not None:
        mod_b = mod_b.attach_modifiers(runtime.get_heapsize_modifiers(size))
    if fd:
        prologue = get_log_prologue(runtime, mod_b)
        fd.write(prologue.encode("ascii"))
    output, companion_out, exit_status = mod_b.run(runtime, cwd=runbms_dir)
    if fd:
        fd.write(output)
        if companion_out:
            fd.write(b"*****\n")
            fd.write(companion_out)
    if fd:
        epilogue = get_log_epilogue(runtime, mod_b)
        fd.write(epilogue.encode("ascii"))
    return output, exit_status


def get_filename_no_ext(
    bm: Benchmark, hfac: Optional[float], size: Optional[int], config: str
) -> str:
    # If we have / in benchmark names, replace with -.
    safe_bm_name = bm.name.replace("/", "-")
    return "{}.{}.{}.{}.{}".format(
        safe_bm_name,
        # plotty uses "^(\w+)\.(\d+)\.(\d+)\.([a-zA-Z0-9_\-\.\:\,]+)\.log\.gz$"
        # to match filenames
        hfac_str(hfac) if hfac is not None else "0",
        size if size is not None else "0",
        config_str_encode(config),
        bm.suite_name,
    )


def get_filename(
    bm: Benchmark, hfac: Optional[float], size: Optional[int], config: str
) -> str:
    return get_filename_no_ext(bm, hfac, size, config) + ".log"


def get_filename_completed(
    bm: Benchmark, hfac: Optional[float], size: Optional[int], config: str
) -> str:
    return "{}.gz".format(get_filename(bm, hfac, size, config))


def get_log_epilogue(runtime: Runtime, bm: Benchmark) -> str:
    return ""


def hz_to_ghz(hzstr: str) -> str:
    return "{:.2f} GHz".format(int(hzstr) / 1000 / 1000)


def get_log_prologue(runtime: Runtime, bm: Benchmark) -> str:
    output = "\n-----\n"
    output += "mkdir -p PLOTTY_WORKAROUND; timedrun; "
    output += bm.to_string(runtime)
    output += "\n"
    output += "running-ng v{}\n".format(__VERSION__)
    output += system("date") + "\n"
    output += system("w") + "\n"
    output += system("vmstat 1 2") + "\n"
    output += system("top -bcn 1 -w512 |head -n 12") + "\n"
    output += "Environment variables: \n"
    for k, v in sorted(os.environ.items()):
        output += "\t{}={}\n".format(k, v)
    output += "OS: "
    output += system("uname -a")
    output += "CPU: "
    output += system("cat /proc/cpuinfo | grep 'model name' | head -1")
    output += "number of cores: "
    cores = system("cat /proc/cpuinfo | grep MHz | wc -l")
    output += cores
    has_cpufreq = Path("/sys/devices/system/cpu/cpu0/cpufreq").is_dir()
    if has_cpufreq:
        for i in range(0, int(cores)):
            output += "Frequency of cpu {}: ".format(i)
            output += hz_to_ghz(
                system(
                    "cat /sys/devices/system/cpu/cpu{}/cpufreq/scaling_cur_freq".format(
                        i
                    )
                )
            )
            output += "\n"
            output += "Governor of cpu {}: ".format(i)
            output += system(
                "cat /sys/devices/system/cpu/cpu{}/cpufreq/scaling_governor".format(i)
            )
            output += "Scaling_min_freq of cpu {}: ".format(i)
            output += hz_to_ghz(
                system(
                    "cat /sys/devices/system/cpu/cpu{}/cpufreq/scaling_min_freq".format(
                        i
                    )
                )
            )
            output += "\n"
    return output


def run_one_benchmark(
    invocations: int,
    suite: BenchmarkSuite,
    bm: Benchmark,
    hfac: Optional[float],
    configs: List[str],
    runbms_dir: Path,
    log_dir: Path,
):
    p: "RunbmsPlugin"
    bm_name = bm.name
    print(bm_name, end=" ")
    size: Optional[int]  # heap size measured in MB
    if hfac is not None:
        print(hfac_str(hfac), end=" ")
        size = get_heapsize(hfac, suite.get_minheap(bm))
        print(size, end=" ")
    else:
        size = None
    for p in plugins.values():
        p.start_benchmark(hfac, size, bm)
    oomed_count: DefaultDict[str, int]
    oomed_count = DefaultDict(int)
    timeout_count: DefaultDict[str, int]
    timeout_count = DefaultDict(int)
    logged_in_users: Set[str]
    logged_in_users = get_logged_in_users()
    if len(logged_in_users) > 1:
        logging.warning(
            "More than one user logged in: {}".format(" ".join(logged_in_users))
        )
    ever_ran = [False] * len(configs)
    for i in range(0, invocations):
        for p in plugins.values():
            p.start_invocation(hfac, size, bm, i)
        print(i, end="", flush=True)
        for j, c in enumerate(configs):
            config_passed = False
            for p in plugins.values():
                p.start_config(hfac, size, bm, i, c, j)
            if skip_oom is not None and oomed_count[c] >= skip_oom:
                print(".", end="", flush=True)
                continue
            if skip_timeout is not None and timeout_count[c] >= skip_timeout:
                print(".", end="", flush=True)
                continue
            if resume:
                log_filename_completed = get_filename_completed(bm, hfac, size, c)
                if (log_dir / log_filename_completed).exists():
                    print(config_index_to_chr(j), end="", flush=True)
                    continue
            log_filename = get_filename(bm, hfac, size, c)
            logging.debug("Running with log filename {}".format(log_filename))
            runtime, _ = parse_config_str(configuration, c)
            if is_dry_run():
                output, exit_status = run_benchmark_with_config(
                    c, bm, runbms_dir, size, None
                )
                assert exit_status is SubprocessrExit.Dryrun
            else:
                fd: BinaryIO
                with (log_dir / log_filename).open("ab") as fd:
                    output, exit_status = run_benchmark_with_config(
                        c, bm, runbms_dir, size, fd
                    )
                ever_ran[j] = True
            if runtime.is_oom(output):
                oomed_count[c] += 1
            if exit_status is SubprocessrExit.Timeout:
                timeout_count[c] += 1
                print(".", end="", flush=True)
            elif exit_status is SubprocessrExit.Error:
                print(".", end="", flush=True)
            elif exit_status is SubprocessrExit.Normal:
                if suite.is_passed(output):
                    config_passed = True
                    print(config_index_to_chr(j), end="", flush=True)
                else:
                    print(".", end="", flush=True)
            elif exit_status is SubprocessrExit.Dryrun:
                print(".", end="", flush=True)
            else:
                raise ValueError("Not a valid SubprocessrExit value")
            for p in plugins.values():
                p.end_config(hfac, size, bm, i, c, j, config_passed)

        for p in plugins.values():
            p.end_invocation(hfac, size, bm, i)
    for p in plugins.values():
        p.end_benchmark(hfac, size, bm)
    for j, c in enumerate(configs):
        log_filename = get_filename(bm, hfac, size, c)
        # Check that this is not a dry-run and we have actually executed this
        # config for a particular benchmark/hfac (method parameters)
        if not is_dry_run() and ever_ran[j]:
            if not skip_log_compression:
                subprocess.check_call(["gzip", log_dir / log_filename])
    print()


def run_one_hfac(
    invocations: int,
    hfac: Optional[float],
    suites: Dict[str, BenchmarkSuite],
    benchmarks: Dict[str, List[Benchmark]],
    configs: List[str],
    runbms_dir: Path,
    log_dir: Path,
):
    p: "RunbmsPlugin"
    for p in plugins.values():
        p.start_hfac(hfac)
    for suite_name, bms in benchmarks.items():
        suite = suites[suite_name]
        for bm in bms:
            run_one_benchmark(
                invocations, suite, bm, hfac, configs, runbms_dir, log_dir
            )
            rsync(log_dir)
    for p in plugins.values():
        p.end_hfac(hfac)


def ensure_remote_dir(log_dir):
    if not is_dry_run() and remote_host is not None:
        log_dir = log_dir.resolve()
        system("ssh {} mkdir -p {}".format(remote_host, log_dir))


def rsync(log_dir):
    if not is_dry_run() and remote_host is not None:
        log_dir = log_dir.resolve()
        system("rsync -ae ssh {}/ {}:{}".format(log_dir, remote_host, log_dir))


def run(args):
    if args.get("which") != "runbms":
        return False
    with tempfile.TemporaryDirectory(prefix="runbms-") as runbms_dir:
        logging.info("Temporary directory: {}".format(runbms_dir))
        if args.get("workdir"):
            args.get("workdir").mkdir(parents=True, exist_ok=True)
            runbms_dir = str(args.get("workdir").resolve())
        # Processing command lines args
        global resume
        resume = args.get("resume")
        if resume:
            run_id = resume
        else:
            prefix = args.get("id_prefix")
            run_id = getid()
            if prefix:
                run_id = "{}-{}".format(prefix, run_id)
        print("Run id: {}".format(run_id))
        log_dir = args.get("LOG_DIR") / run_id
        if not is_dry_run():
            log_dir.mkdir(parents=True, exist_ok=True)
            with (log_dir / "runbms_args.yml").open("w") as fd:
                yaml.dump(args, fd)
        N = args.get("N")
        ns = args.get("n")
        slice = args.get("slice")
        slice = [float(s) for s in slice.split(",")] if slice else []
        global skip_oom
        skip_oom = args.get("skip_oom")
        global skip_timeout
        skip_timeout = args.get("skip_timeout")
        global skip_log_compression
        skip_log_compression = args.get("skip_log_compression")
        # Load from configuration file
        global configuration
        configuration = Configuration.from_file(Path(os.getcwd()), args.get("CONFIG"))
        # Save metadata
        if not is_dry_run():
            with (log_dir / "runbms.yml").open("w") as fd:
                configuration.save_to_file(fd)
        configuration.resolve_class()
        # Read from configuration, override with command line arguments if
        # needed
        invocations = configuration.get("invocations")
        if args.get("invocations"):
            invocations = args.get("invocations")
        global minheap_multiplier
        minheap_multiplier = configuration.get("minheap_multiplier")
        if args.get("minheap_multiplier"):
            minheap_multiplier = args.get("minheap_multiplier")
        heap_range = configuration.get("heap_range")
        spread_factor = configuration.get("spread_factor")
        suites = configuration.get("suites")
        benchmarks = configuration.get("benchmarks")
        if benchmarks is None:
            benchmarks = {}
        configs = configuration.get("configs")
        global remote_host
        remote_host = configuration.get("remote_host")
        if not is_dry_run() and remote_host is not None:
            ensure_remote_dir(log_dir)
        global plugins
        plugins = configuration.get("plugins")
        if plugins is None:
            plugins = {}
        else:
            from running.plugin.runbms import RunbmsPlugin

            if type(plugins) is not dict:
                raise TypeError("plugins must be a dictionary")
            plugins = {k: RunbmsPlugin.from_config(k, v) for k, v in plugins.items()}
            for p in plugins.values():
                p.set_run_id(run_id)
                p.set_runbms_dir(runbms_dir)
                p.set_log_dir(log_dir)

        def run_hfacs(hfacs):
            logging.info(
                "hfacs: {}".format(", ".join([hfac_str(hfac) for hfac in hfacs]))
            )
            for hfac in hfacs:
                run_one_hfac(
                    invocations,
                    hfac,
                    suites,
                    benchmarks,
                    configs,
                    Path(runbms_dir),
                    log_dir,
                )
                print()

        def run_N_ns(N, ns):
            hfacs = get_hfacs(heap_range, spread_factor, N, ns)
            run_hfacs(hfacs)

        if slice:
            run_hfacs(slice)
            return True

        if N is None:
            run_one_hfac(
                invocations,
                None,
                suites,
                benchmarks,
                configs,
                Path(runbms_dir),
                log_dir,
            )
            return True

        if len(ns) == 0:
            fillin(run_N_ns, round(math.log2(N)))
        else:
            run_N_ns(N, ns)

        return True



from typing import Any, Dict
import yaml
from running.suite import BenchmarkSuite
from running.runtime import Runtime
from running.modifier import Modifier
from pathlib import Path
import functools
import copy
import logging
import os


def load_class(cls, config):
    return {k: cls.from_config(k, v) for (k, v) in config.items()}


KEY_CLASS_MAPPING = {
    "suites": BenchmarkSuite,
    "modifiers": Modifier,
    "runtimes": Runtime,
}


class Configuration(object):
    def __init__(self, kv_pairs: Dict[str, Any]):
        assert "includes" not in kv_pairs
        assert "overrides" not in kv_pairs
        self.__items = kv_pairs

    def save_to_file(self, fd):
        yaml.dump(self.__items, fd)

    def resolve_class(self):
        """Resolve the values by instantiating instances of classes

        For example, self.values["suites"] is a Dict[str, Dict[str, str]],
        where in the inner dictionary contains the string representation of a
        benchmark suite.
        After this function returns, self.values["suites"] becomes a
        Dict[str, BenchmarkSuite].

        Change the KEY_CLASS_MAPPING to change which classes get resolved.
        """
        for cls_name, cls in KEY_CLASS_MAPPING.items():
            if cls_name in self.__items:
                self.__items[cls_name] = load_class(cls, self.__items[cls_name])
        if "benchmarks" in self.__items:
            for suite_name, bms in self.__items["benchmarks"].items():
                suite = self.__items["suites"][suite_name]
                benchmarks = []
                for b in bms:
                    benchmarks.append(suite.get_benchmark(b))
                self.__items["benchmarks"][suite_name] = benchmarks

    def get(self, name: str) -> Any:
        return self.__items.get(name)

    def override(self, selector: str, new_value: Any):
        current: Any  # Union[Dict[str, Any], List[Any]]
        current = self.__items
        parts = list(selector.split("."))
        for index, p in enumerate(parts):
            if index == len(parts) - 1:
                if p.isnumeric():
                    current[int(p)] = new_value
                else:
                    current[p] = new_value
            else:
                if p.isnumeric():
                    current = current[int(p)]
                else:
                    current = current[p]

    def combine(self, other: "Configuration") -> "Configuration":
        """Combine top-level items of self.values.

        Arrays are concatenated and dictionaries are updated.
        """
        new_values = copy.deepcopy(self.__items)
        for k, v in other.__items.items():
            if k in new_values:
                if type(new_values[k]) is list:
                    new_values[k].extend(copy.deepcopy(other.__items[k]))
                else:
                    if type(new_values[k]) is not dict:
                        raise TypeError(
                            "Key `{}` has been defined in one of the "
                            "included files, and the value of `{}`, {}, "
                            "is not an array or a dictionary. "
                            "Please use overrides instead.".format(k, k, repr(v))
                        )
                    new_values[k].update(copy.deepcopy(other.__items[k]))
            else:
                new_values[k] = copy.deepcopy(other.__items[k])
        return Configuration(new_values)

    @staticmethod
    def parse_file(path: Path) -> Any:
        with path.open("r") as fd:
            try:
                config = yaml.safe_load(fd)
                return config
            except yaml.YAMLError as e:
                raise SyntaxError(
                    "Not able to parse the configuration file, {}".format(e)
                )

    @staticmethod
    def from_file(in_folder: Path, p: str) -> "Configuration":
        expand_p = os.path.expandvars(p)
        logging.info(
            "Loading config {}, expanding to {}, relative to {}".format(
                p, expand_p, in_folder
            )
        )
        path = Path(expand_p)
        if path.is_absolute():
            logging.info("    is absolute")
        else:
            path = in_folder.joinpath(p)
            logging.info("    resolved to {}".format(path))
        if not path.exists():
            raise ValueError("Configuration not found at path '{}'".format(path))
        if not path.is_file():
            raise ValueError("Configuration at path '{}' is not a file".format(path))
        with path.open("r") as fd:
            try:
                config = yaml.safe_load(fd)
            except yaml.YAMLError as e:
                raise SyntaxError(
                    "Not able to parse the configuration file, {}".format(e)
                )
        if config is None:
            raise ValueError("Parsed configuration file is None")
        if "includes" in config:
            includes = [
                Configuration.from_file(path.parent, p) for p in config["includes"]
            ]
            base = functools.reduce(lambda left, right: left.combine(right), includes)
            if "overrides" in config:
                for selector, new_value in config["overrides"].items():
                    base.override(selector, new_value)
                del config["overrides"]
            del config["includes"]
            final_config = Configuration(config)
            final_config = base.combine(final_config)
        else:
            if "overrides" in config:
                raise KeyError(
                    'You specified "overrides" but not "includes". This does not make sense.'
                )
            final_config = Configuration(config)
        return final_config



from running.config import Configuration


def test_openjdk_path_ennvvar():
    c = Configuration(
        {
            "runtimes": {
                "temurin-21_bogus": {
                    "type": "OpenJDK",
                    "release": 21,
                    # some bogus environment variable that will not be expanded
                    "home": "$DAHKDLHDIWHEIUWHEIWEHIJHDJKAGDKJADGUQDGIQUWDGI/temurin-21-jdk-amd64",
                },
                "temurin-21": {
                    "type": "OpenJDK",
                    "release": 21,
                    "home": "$HOME/temurin-21-jdk-amd64",
                },
            }
        }
    )

    c.resolve_class()
    temurin_21_bogus = c.get("runtimes")["temurin-21_bogus"]
    temurin_21 = c.get("runtimes")["temurin-21_bogus"]
    assert "$HOME" not in str(temurin_21.home)
    assert "$DAHKDLHDIWHEIUWHEIWEHIJHDJKAGDKJADGUQDGIQUWDGI" in str(
        temurin_21_bogus.home
    )



from typing import Any, Dict, List, TYPE_CHECKING
from running.util import register, smart_quote, split_quoted, parse_modifier_strs
import copy

if TYPE_CHECKING:
    from running.config import Configuration


class Modifier(object):
    CLS_MAPPING: Dict[str, Any]
    CLS_MAPPING = {}

    def __init__(self, value_opts=None, **kwargs):
        self.name = kwargs["name"]
        self.value_opts = value_opts
        if "-" in self.name:
            raise ValueError(
                "Modifier {} has - in its name. - is reserved for value options.".format(
                    self.name
                )
            )
        self.__original_kwargs = kwargs
        self._kwargs = copy.deepcopy(kwargs)
        self.excludes = kwargs.get("excludes", {})
        self.includes = kwargs.get("includes", {})
        if self.value_opts:  # Neither None nor empty
            # Expand value opts
            for k, v in kwargs.items():
                if type(v) is not str:
                    continue
                try:
                    self._kwargs[k] = v.format(*value_opts)
                except IndexError:
                    pass

    @staticmethod
    def from_config(name: str, config: Dict[str, str]) -> Any:
        return Modifier.CLS_MAPPING[config["type"]](name=name, **config)

    def apply_value_opts(self, value_opts):
        return type(self)(value_opts=value_opts, **self.__original_kwargs)

    def should_attach(self, suite_name: str, bm_name: str):
        # We only attach this modifier to the benchmark listed in the
        # includes list
        if self.includes:
            if suite_name not in self.includes:
                return False
            if bm_name not in self.includes[suite_name]:
                return False
        # If a benchmark is in the excludes list, then we don't attach
        # the modifier
        if suite_name in self.excludes and bm_name in self.excludes[suite_name]:
            return False
        return True

    def __str__(self) -> str:
        return "Modifier {}".format(self.name)


@register(Modifier)
class ModifierSet(Modifier):
    def __init__(self, value_opts=None, **kwargs):
        super().__init__(value_opts, **kwargs)
        self.val = self._kwargs["val"].split("|")

    def flatten(self, configuration: "Configuration") -> List[Modifier]:
        return parse_modifier_strs(configuration, self.val)

    def __str__(self) -> str:
        return "{} ModifierSet {}".format(super().__str__(), "|".join(self.val))


@register(Modifier)
class JVMArg(Modifier):
    def __init__(self, value_opts=None, **kwargs):
        super().__init__(value_opts, **kwargs)
        self.val = split_quoted(self._kwargs["val"])

    def __str__(self) -> str:
        return "{} JVMArg {}".format(super().__str__(), self.val)


@register(Modifier)
class JVMClasspathAppend(Modifier):
    def __init__(self, value_opts=None, **kwargs):
        super().__init__(value_opts, **kwargs)
        self.val = split_quoted(self._kwargs["val"])

    def __str__(self) -> str:
        return "{} JVMClasspathAppend {}".format(super().__str__(), self.val)


@register(Modifier)
class JVMClasspath(JVMClasspathAppend):
    # backward compatibility
    pass


@register(Modifier)
class JVMClasspathPrepend(Modifier):
    def __init__(self, value_opts=None, **kwargs):
        super().__init__(value_opts, **kwargs)
        self.val = split_quoted(self._kwargs["val"])

    def __str__(self) -> str:
        return "{} JVMClasspathPrepend {}".format(super().__str__(), self.val)


@register(Modifier)
class EnvVar(Modifier):
    def __init__(self, value_opts=None, **kwargs):
        super().__init__(value_opts, **kwargs)
        if "var" not in self._kwargs:
            raise ValueError(
                "Please specify the name of the environment variable for modifier {}".format(
                    self.name
                )
            )
        if "val" not in self._kwargs:
            raise ValueError(
                "Please specify the value for the environment variable for modifier {}".format(
                    self.name
                )
            )
        self.var = self._kwargs["var"]
        self.val = self._kwargs["val"]

    def __str__(self) -> str:
        return "{} EnvVar {}={}".format(
            super().__str__(), self.var, smart_quote(self.val)
        )


@register(Modifier)
class ProgramArg(Modifier):
    def __init__(self, value_opts=None, **kwargs):
        super().__init__(value_opts, **kwargs)
        self.val = split_quoted(self._kwargs["val"])

    def __str__(self) -> str:
        return "{} ProgramArg {}".format(super().__str__(), self.val)


@register(Modifier)
class Wrapper(Modifier):
    def __init__(self, value_opts=None, **kwargs):
        super().__init__(value_opts, **kwargs)
        self.val = split_quoted(self._kwargs["val"])

    def __str__(self) -> str:
        return "{} Wrapper {}".format(super().__str__(), self.val)


@register(Modifier)
class JSArg(Modifier):
    def __init__(self, value_opts=None, **kwargs):
        super().__init__(value_opts, **kwargs)
        self.val = split_quoted(self._kwargs["val"])

    def __str__(self) -> str:
        return "{} JSArg {}".format(super().__str__(), self.val)


@register(Modifier)
class Companion(Modifier):
    def __init__(self, value_opts=None, **kwargs):
        super().__init__(value_opts, **kwargs)
        self.val = split_quoted(self._kwargs["val"])

    def __str__(self) -> str:
        return "{} Companion {}".format(super().__str__(), self.val)


@register(Modifier)
class JuliaArg(Modifier):
    def __init__(self, value_opts=None, **kwargs):
        super().__init__(value_opts, **kwargs)
        self.val = split_quoted(self._kwargs["val"])

    def __str__(self) -> str:
        return "{} JuliaArg {}".format(super().__str__(), self.val)



from running.config import Configuration
from running.runtime import DummyRuntime
from running.suite import BinaryBenchmarkSuite


def test_binary_benchmark_suite_quoted():
    b = BinaryBenchmarkSuite(
        name="foobar", programs={"ls": {"path": "/bin/ls", "args": 'foo "bar buzz"'}}
    )
    assert b.programs["ls"]["args"] == ["foo", "bar buzz"]


def test_dacapo_size():
    c = Configuration(
        {
            "suites": {
                "dacapo2006": {
                    "type": "DaCapo",
                    "release": "2006",
                    "path": "/usr/share/benchmarks/dacapo/dacapo-2006-10-MR2.jar",
                    "timing_iteration": 3,
                },
                "dacapo2006_default": {
                    "type": "DaCapo",
                    "release": "2006",
                    "path": "/usr/share/benchmarks/dacapo/dacapo-2006-10-MR2.jar",
                    "timing_iteration": 3,
                    "size": "default",
                },
            },
            "benchmarks": {
                "dacapo2006": [
                    "fop",
                    dict(name="fop_small", bm_name="fop", size="small"),
                ],
                "dacapo2006_default": ["fop"],
            },
        }
    )

    c.resolve_class()
    fop = c.get("benchmarks")["dacapo2006"][0]
    fop_default = c.get("benchmarks")["dacapo2006_default"][0]
    fop_small = c.get("benchmarks")["dacapo2006"][1]
    assert "-s" not in fop.to_string(DummyRuntime("java"))
    assert "-s default" in fop_default.to_string(DummyRuntime("java"))
    assert "-s small" in fop_small.to_string(DummyRuntime("java"))


def test_dacapo_timing_iteration():
    c = Configuration(
        {
            "suites": {
                "dacapo2006": {
                    "type": "DaCapo",
                    "release": "2006",
                    "path": "/usr/share/benchmarks/dacapo/dacapo-2006-10-MR2.jar",
                    "timing_iteration": 3,
                }
            },
            "benchmarks": {
                "dacapo2006": [
                    "fop",
                    dict(
                        name="fop_converge", bm_name="fop", timing_iteration="converge"
                    ),
                ]
            },
        }
    )

    c.resolve_class()
    fop = c.get("benchmarks")["dacapo2006"][0]
    fop_converge = c.get("benchmarks")["dacapo2006"][1]
    assert "-n 3" in fop.to_string(DummyRuntime("java"))
    assert "-converge" in fop_converge.to_string(DummyRuntime("java"))


def test_dacapo_path_ennvvar():
    c = Configuration(
        {
            "suites": {
                "dacapo2006_bogus": {
                    "type": "DaCapo",
                    "release": "2006",
                    # some bogus environment variable that will not be expanded
                    "path": "$DAHKDLHDIWHEIUWHEIWEHIJHDJKAGDKJADGUQDGIQUWDGI/dacapo-2006-10-MR2.jar",
                    "timing_iteration": 3,
                },
                "dacapo2006": {
                    "type": "DaCapo",
                    "release": "2006",
                    "path": "$HOME/dacapo-2006-10-MR2.jar",
                    "timing_iteration": 3,
                },
            }
        }
    )

    c.resolve_class()
    dacapo2006 = c.get("suites")["dacapo2006"]
    dacapo2006_bogus = c.get("suites")["dacapo2006_bogus"]
    assert "$HOME" not in str(dacapo2006.path)
    assert "$DAHKDLHDIWHEIUWHEIWEHIJHDJKAGDKJADGUQDGIQUWDGI" in str(
        dacapo2006_bogus.path
    )



from running.config import Configuration
import pytest


def test_override():
    c = Configuration({"a": {"b": 1, "c": 42}, "d": ["foo", "bar"]})
    c.override("a.c", 43)
    c.override("d.1", "buzz")
    assert c.get("a")["b"] == 1
    assert c.get("a")["c"] == 43
    assert c.get("d") == ["foo", "buzz"]


def test_combine():
    c1 = Configuration({"a": {"b": 1, "c": 42}, "d": ["foo", "bar"]})

    c2 = Configuration({"a": {"b": 2, "e": 43}, "d": ["fizz", "buzz"], "f": 100})

    c = c1.combine(c2)
    assert c.get("a") == {"b": 2, "c": 42, "e": 43}
    assert c.get("d") == ["foo", "bar", "fizz", "buzz"]
    assert c.get("f") == 100


def test_combine_fail():
    c1 = Configuration({"a": "val1", "b": "b"})

    c2 = Configuration({"a": "val2", "c": "c"})

    with pytest.raises(TypeError):
        c1.combine(c2)


def test_resolve_suites():
    c = Configuration(
        {
            "suites": {
                "dacapo2006": {
                    "type": "DaCapo",
                    "release": "2006",
                    "path": "/usr/share/benchmarks/dacapo/dacapo-2006-10-MR2.jar",
                    "timing_iteration": 3,
                }
            }
        }
    )
    c.resolve_class()
    dacapo2006 = c.get("suites")["dacapo2006"]
    assert dacapo2006.release == "2006"
    assert dacapo2006.path.stem == "dacapo-2006-10-MR2"


def test_resolve_modifiers():
    c = Configuration(
        {
            "modifiers": {
                "ss": {"type": "EnvVar", "var": "MMTK_PLAN", "val": "SemiSpace"}
            }
        }
    )
    c.resolve_class()
    ss = c.get("modifiers")["ss"]
    assert ss.var == "MMTK_PLAN"
    assert ss.val == "SemiSpace"


def test_resolve_jvms():
    c = Configuration(
        {
            "runtimes": {
                "adoptopenjdk-8": {
                    "type": "OpenJDK",
                    "release": "8",
                    "home": "/usr/lib/jvm/adoptopenjdk-8-hotspot-amd64",
                }
            }
        }
    )
    c.resolve_class()
    jdk8 = c.get("runtimes")["adoptopenjdk-8"]
    assert str(jdk8.executable) == "/usr/lib/jvm/adoptopenjdk-8-hotspot-amd64/bin/java"
    assert jdk8.release == 8



from pathlib import Path
from typing import Any, Dict, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from running.benchmark import Benchmark


class RunbmsPlugin(object):
    CLS_MAPPING: Dict[str, Any]
    CLS_MAPPING = {}

    def __init__(self, **kwargs):
        self.name = kwargs["name"]
        self.run_id = None
        self.runbms_dir: Optional[Path]
        self.runbms_dir = None
        self.log_dir: Optional[Path]
        self.log_dir = None

    def set_run_id(self, run_id: Path):
        self.run_id = run_id

    def set_runbms_dir(self, runbms_dir: str):
        self.runbms_dir = Path(runbms_dir).resolve()

    def set_log_dir(self, log_dir: Path):
        self.log_dir = log_dir

    @staticmethod
    def from_config(name: str, config: Dict[str, str]) -> Any:
        return RunbmsPlugin.CLS_MAPPING[config["type"]](name=name, **config)

    def __str__(self) -> str:
        return "RunbmsPlugin {}".format(self.name)

    def start_hfac(self, _hfac: Optional[float]):
        pass

    def end_hfac(self, _hfac: Optional[float]):
        pass

    def start_benchmark(
        self, _hfac: Optional[float], _size: Optional[int], _bm: "Benchmark"
    ):
        pass

    def end_benchmark(
        self, _hfac: Optional[float], _size: Optional[int], _bm: "Benchmark"
    ):
        pass

    def start_invocation(
        self,
        _hfac: Optional[float],
        _size: Optional[int],
        _bm: "Benchmark",
        _invocation: int,
    ):
        pass

    def end_invocation(
        self,
        _hfac: Optional[float],
        _size: Optional[int],
        _bm: "Benchmark",
        _invocation: int,
    ):
        pass

    def start_config(
        self,
        _hfac: Optional[float],
        _size: Optional[int],
        _bm: "Benchmark",
        _invocation: int,
        _config: str,
        _config_index: int,
    ):
        pass

    def end_config(
        self,
        _hfac: Optional[float],
        _size: Optional[int],
        _bm: "Benchmark",
        _invocation: int,
        _config: str,
        _config_index: int,
        _passed: bool,
    ):
        pass


# !!! Do NOT remove this import nor change its position
# This is to make sure that the plugin classes are correctly registered
from running.plugin.runbms.copyfile import CopyFile

if TYPE_CHECKING:
    from running.plugin.runbms.zulip import Zulip
else:
    try:
        from running.plugin.runbms.zulip import Zulip
    except:
        from running.util import register

        @register(RunbmsPlugin)
        class Zulip(RunbmsPlugin):
            def __init__(self, **kwargs):
                raise RuntimeError(
                    "Trying to create an instance of the Zulip "
                    "plugin for runbms, but the import failed. "
                    "This is most likely due to the required "
                    "dependencies not being installed. Try pip "
                    "install running-ng[zulip] to install the extra dependencies."
                )



from running.command.fillin import fillin


def test_fillin():
    results = []
    fillin(lambda base, ns: results.append(list(ns)), 3)
    assert results == [[0, 4, 8], [2, 6], [1, 3, 5, 7]]

    results = []
    fillin(lambda base, ns: results.append(list(ns)), 4)
    assert results == [[0, 8, 16], [4, 12], [2, 6, 10, 14], [1, 3, 5, 7, 9, 11, 13, 15]]


def test_fillin_start():
    results = []
    fillin(lambda base, ns: results.append(list(ns)), 3, 1)
    assert results == [[1, 3, 5, 7]]

    results = []
    fillin(lambda base, ns: results.append(list(ns)), 4, 2)
    assert results == [[2, 6, 10, 14], [1, 3, 5, 7, 9, 11, 13, 15]]



from typing import Optional, Iterable, Callable
import subprocess


def fillin(
    callback: Callable[[int, Iterable[int]], None],
    levels: int,
    start: Optional[int] = None,
):
    """Fill the parameter space

    The parameter space is from 0, 1, 2, ..., 2^levels (not right-inclusive).

    The advantage of using this function is that it fills in the points in the
    space logarithmically, starting with the ends and middle, and progressing
    until all 2^levels + 1 points are explored.

    Parameters
    ----------
    callback : Callable[[int, Iterable[int]], None]
        The arguments represent a list of fractions.
        For example, callback(8, [2, 6]) means 2/8 and 6/8.
    levels : int
        The log value of the size of the parameter space.
    start: Optional[int]
        Instead of starting from 0, start from a specified value.
    """
    commenced = False
    if start is None:
        callback(2**levels, range(0, 2**levels + 1, 2 ** (levels - 1)))
        commenced = True
    i = 1
    while i < levels:
        base = 2 ** (levels - 1 - i)  # larger i, smaller base
        step = 2 ** (levels - i)  # larger i, smaller step
        if start is not None and base == start:
            commenced = True
        if commenced:
            callback(2**levels, range(base, 2**levels, step))
        i += 1


def cmd_program(prog) -> Callable[[int, Iterable[int]], None]:
    def callback(end, ns):
        cmd = [prog]
        cmd.append(str(end))
        cmd.extend(map(str, ns))
        output = subprocess.check_output(cmd)
        print(output.decode("utf-8"), end="")

    return callback


def setup_parser(subparsers):
    f = subparsers.add_parser("fillin")
    f.set_defaults(which="fillin")
    f.add_argument("PROG")
    f.add_argument("LEVELS", type=int)
    f.add_argument("START", type=int, nargs="?", default=None)


def run(args):
    if args.get("which") != "fillin":
        return False
    callback = cmd_program(args.get("PROG"))
    fillin(callback, args.get("LEVELS"), args.get("START"))
    return True



from running.command.log_preprocessor import (
    filter_stats,
    reduce_stats,
    sum_work_perf_event,
    calc_work_ipc,
    ratio_work_perf_event,
)


def test_filter():
    assert filter_stats(lambda x: "foo" in x)(
        {"foo1": 1, "2foo": 2, "bar3": 3, "4bar": 4}
    ) == {"foo1": 1, "2foo": 2}


def test_reduce():
    assert (
        reduce_stats(r"\d*foo\d*", "foo_sum", lambda x, y: x + y)(
            {"foo1": 1, "2foo": 2, "bar3": 3, "4bar": 4}
        )["foo_sum"]
        == 3
    )


def test_sum_perf():
    assert (
        sum_work_perf_event("PERF_COUNT_HW_CPU_CYCLES")(
            {
                "work.ScanUniverseRoots.PERF_COUNT_HW_CPU_CYCLES.total": 1,
                "work.PrepareCollector.PERF_COUNT_HW_CPU_CYCLES.total": 2,
                "work.ScanUniverseRoots.PERF_COUNT_HW_CACHE_L1D:MISS.total": 3,
            }
        )["work.PERF_COUNT_HW_CPU_CYCLES.total"]
        == 3
    )


def test_ratio_perf():
    stats = {
        "work.ScanUniverseRoots.PERF_COUNT_HW_CPU_CYCLES.total": 2,
        "work.PrepareCollector.PERF_COUNT_HW_CPU_CYCLES.total": 3,
        "work.foo.PERF_COUNT_HW_CPU_CYCLES.total": 5,
    }
    stats = sum_work_perf_event("PERF_COUNT_HW_CPU_CYCLES")(stats)
    ratios = ratio_work_perf_event("PERF_COUNT_HW_CPU_CYCLES")(stats)
    assert ratios["work.ScanUniverseRoots.PERF_COUNT_HW_CPU_CYCLES.ratio"] == 0.2
    assert ratios["work.PrepareCollector.PERF_COUNT_HW_CPU_CYCLES.ratio"] == 0.3
    assert ratios["work.foo.PERF_COUNT_HW_CPU_CYCLES.ratio"] == 0.5


def test_ipc():
    ipcs = calc_work_ipc(
        {
            "work.foo.PERF_COUNT_HW_CPU_CYCLES.total": 20,
            "work.foo.PERF_COUNT_HW_INSTRUCTIONS.total": 4,
            "work.bar.PERF_COUNT_HW_CPU_CYCLES.total": 10,
            "work.bar.PERF_COUNT_HW_INSTRUCTIONS.total": 5,
        }
    )
    print(ipcs)
    assert ipcs["work.foo.INSTRUCTIONS_PER_CYCLE.ratio"] == 0.2
    assert ipcs["work.bar.INSTRUCTIONS_PER_CYCLE.ratio"] == 0.5



import os
import stat
from typing import List, Optional, TYPE_CHECKING
from running.plugin.runbms import RunbmsPlugin
from running.command.runbms import get_filename_no_ext
from running.suite import is_dry_run
from running.util import register
import shutil

if TYPE_CHECKING:
    from running.benchmark import Benchmark


def delete_readonly(_function, path, _excinfo):
    os.chmod(path, stat.S_IWRITE)
    os.remove(path)


@register(RunbmsPlugin)
class CopyFile(RunbmsPlugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.nop: bool
        self.nop = is_dry_run()
        self.patterns: List[str]
        self.patterns = kwargs.get("patterns", [])
        if type(self.patterns) is not list:
            raise TypeError("patterns of CopyFile must be a list")
        self.skip_failed = kwargs.get("skip_failed", True)
        if type(self.skip_failed) is not bool:
            raise TypeError("skip_failed of CopyFile must be a bool")

    def __str__(self) -> str:
        return "CopyFile {}".format(self.name)

    def start_hfac(self, hfac: Optional[float]):
        if self.nop:
            return

    def end_hfac(self, _hfac: Optional[float]):
        if self.nop:
            return

    def start_benchmark(
        self, _hfac: Optional[float], _size: Optional[int], bm: "Benchmark"
    ):
        if self.nop:
            return

    def end_benchmark(
        self, _hfac: Optional[float], _size: Optional[int], bm: "Benchmark"
    ):
        if self.nop:
            return

    def start_invocation(
        self,
        _hfac: Optional[float],
        _size: Optional[int],
        _bm: "Benchmark",
        invocation: int,
    ):
        if self.nop:
            return

    def end_invocation(
        self,
        _hfac: Optional[float],
        _size: Optional[int],
        _bm: "Benchmark",
        _invocation: int,
    ):
        if self.nop:
            return

    def start_config(
        self,
        _hfac: Optional[float],
        _size: Optional[int],
        _bm: "Benchmark",
        _invocation: int,
        config: str,
        _config_index: int,
    ):
        if self.nop:
            return

    def end_config(
        self,
        hfac: Optional[float],
        size: Optional[int],
        bm: "Benchmark",
        invocation: int,
        config: str,
        _config_index: int,
        passed: bool,
    ):
        if self.nop:
            return
        if self.runbms_dir is None:
            raise ValueError("runbms_dir should be set")
        if self.log_dir is None:
            raise ValueError("log_dir should be set")
        folder_name = "{}.{}".format(
            get_filename_no_ext(bm, hfac, size, config), invocation
        )
        if self.skip_failed and (not passed):
            # Do nothing if we skip failed invocation and the current invocation
            # didn't pass
            pass
        else:
            target = self.log_dir / folder_name
            target.mkdir(parents=True, exist_ok=True)
            for pattern in self.patterns:
                for file in self.runbms_dir.glob(pattern):
                    shutil.copy2(file, target)
        for child in self.runbms_dir.iterdir():
            if child.is_dir():
                shutil.rmtree(child)
            else:
                child.unlink()



import logging
import subprocess
import sys
from time import sleep
from typing import Callable, Sequence, TypeVar, List, Optional, Tuple, Union, Dict
from running.runtime import D8, JavaScriptCore, Runtime, DummyRuntime, SpiderMonkey
from running.modifier import *
from running.util import smart_quote, split_quoted
from pathlib import Path
from copy import deepcopy
from running import suite
import os
from enum import Enum

COMPANION_WAIT_START = 2.0


class SubprocessrExit(Enum):
    Normal = 1
    Error = 2
    Timeout = 3
    Dryrun = 4


B = TypeVar("B", bound="Benchmark")


class Benchmark(object):
    def __init__(
        self,
        suite_name: str,
        name: str,
        wrapper: Optional[str] = None,
        timeout: Optional[int] = None,
        override_cwd: Optional[Path] = None,
        companion: Optional[str] = None,
        runtime_specific_modifiers_strategy: Optional[
            Callable[[Runtime], Sequence[Modifier]]
        ] = None,
        **kwargs,
    ):
        self.name = name
        self.suite_name = suite_name
        self.env_args: Dict[str, str]
        self.env_args = {}
        self.wrapper: List[str]
        if wrapper is not None:
            self.wrapper = split_quoted(wrapper)
        else:
            self.wrapper = []
        if companion is not None:
            self.companion = split_quoted(companion)
        else:
            self.companion = []
        self.timeout = timeout
        # ignore the current working directory provided by commands like runbms or minheap
        # certain benchmarks expect to be invoked from certain directories
        self.override_cwd = override_cwd
        self.runtime_specific_modifiers_strategy: Callable[
            [Runtime], Sequence[Modifier]
        ]
        if runtime_specific_modifiers_strategy is not None:
            self.runtime_specific_modifiers_strategy = (
                runtime_specific_modifiers_strategy
            )
        else:
            self.runtime_specific_modifiers_strategy = lambda _runtime: []

    def get_env_str(self) -> str:
        return " ".join(
            [
                "{}={}".format(k, smart_quote(os.path.expandvars(v)))
                for (k, v) in self.env_args.items()
            ]
        )

    def get_full_args(self, _runtime: Runtime) -> List[Union[str, Path]]:
        # makes a copy because the subclass might change the list
        # also to type check https://mypy.readthedocs.io/en/stable/common_issues.html#variance
        return list(self.wrapper)

    def get_runtime_specific_modifiers(self, runtime: Runtime) -> Sequence[Modifier]:
        return self.runtime_specific_modifiers_strategy(runtime)

    def attach_modifiers(self: B, modifiers: Sequence[Modifier]) -> B:
        b = deepcopy(self)
        for m in modifiers:
            if not m.should_attach(self.suite_name, self.name):
                continue
            elif type(m) == Wrapper:
                b.wrapper.extend(m.val)
            elif type(m) == Companion:
                b.companion.extend(m.val)
            elif type(m) == EnvVar:
                b.env_args[m.var] = m.val
            elif type(m) == ModifierSet:
                logging.warning("ModifierSet should have been flattened")
        return b

    def to_string(self, runtime: Runtime) -> str:
        return "{} {}".format(
            self.get_env_str(),
            " ".join(
                [
                    smart_quote(os.path.expandvars(x))
                    for x in self.get_full_args(runtime)
                ]
            ),
        )

    def run(
        self, runtime: Runtime, cwd: Optional[Path] = None
    ) -> Tuple[bytes, bytes, SubprocessrExit]:
        if suite.is_dry_run():
            print(self.to_string(runtime), file=sys.stderr)
            return b"", b"", SubprocessrExit.Dryrun
        else:
            cmd = self.get_full_args(runtime)
            cmd = [os.path.expandvars(x) for x in cmd]
            env_args_to_add = {
                k: os.path.expandvars(v) for k, v in self.env_args.items()
            }
            env_args = os.environ.copy()
            env_args.update(env_args_to_add)
            companion_out = b""
            stdout: Optional[bytes]
            if self.companion:
                companion_p = subprocess.Popen(
                    self.companion, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                )
                sleep(COMPANION_WAIT_START)
            try:
                p = subprocess.run(
                    cmd,
                    env=env_args,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    timeout=self.timeout,
                    cwd=self.override_cwd if self.override_cwd else cwd,
                )
                subprocess_exit = SubprocessrExit.Normal
                stdout = p.stdout
            except subprocess.CalledProcessError as e:
                subprocess_exit = SubprocessrExit.Error
                stdout = e.stdout
            except subprocess.TimeoutExpired as e:
                subprocess_exit = SubprocessrExit.Timeout
                stdout = e.stdout
            finally:
                if self.companion:
                    try:
                        companion_stdout, _ = companion_p.communicate(timeout=10)
                        companion_out += companion_stdout
                    except subprocess.TimeoutExpired:
                        logging.warning(
                            "Companion program not exited after 10 seconds timeout. Trying to kill ..."
                        )
                        try:
                            companion_p.kill()
                        except PermissionError:
                            logging.warning("Failed to kill.")
                        companion_stdout, _ = companion_p.communicate()
                        companion_out += companion_stdout

            return stdout if stdout else b"", companion_out, subprocess_exit


class BinaryBenchmark(Benchmark):
    def __init__(self, program: Path, program_args: List[Union[str, Path]], **kwargs):
        super().__init__(**kwargs)
        self.program = program
        self.program_args = program_args
        assert program.exists()

    def __str__(self) -> str:
        return self.to_string(DummyRuntime(""))

    def attach_modifiers(self, modifiers: Sequence[Modifier]) -> "BinaryBenchmark":
        bb = super().attach_modifiers(modifiers)
        for m in modifiers:
            if not m.should_attach(self.suite_name, self.name):
                continue
            elif type(m) == ProgramArg:
                bb.program_args.extend(m.val)
            elif type(m) == JVMArg:
                logging.warning("JVMArg not respected by BinaryBenchmark")
            elif isinstance(m, JVMClasspathAppend) or type(m) == JVMClasspathPrepend:
                logging.warning("JVMClasspath not respected by BinaryBenchmark")
            elif type(m) == JSArg:
                logging.warning("JSArg not respected by BinaryBenchmark")
        return bb

    def get_full_args(self, _runtime: Runtime) -> List[Union[str, Path]]:
        cmd = super().get_full_args(_runtime)
        cmd.append(self.program)
        cmd.extend(self.program_args)
        return cmd


class JavaBenchmark(Benchmark):
    def __init__(
        self, jvm_args: List[str], program_args: List[str], cp: List[str], **kwargs
    ):
        super().__init__(**kwargs)
        self.jvm_args = jvm_args
        self.program_args = program_args
        self.cp = cp

    def get_classpath_args(self) -> List[str]:
        return ["-cp", ":".join(self.cp)] if self.cp else []

    def __str__(self) -> str:
        return self.to_string(DummyRuntime("java"))

    def attach_modifiers(self, modifiers: Sequence[Modifier]) -> "JavaBenchmark":
        jb = super().attach_modifiers(modifiers)
        for m in modifiers:
            if not m.should_attach(self.suite_name, self.name):
                continue
            if type(m) == JVMArg:
                jb.jvm_args.extend(m.val)
            elif type(m) == ProgramArg:
                jb.program_args.extend(m.val)
            elif isinstance(m, JVMClasspathAppend):
                jb.cp.extend(m.val)
            elif type(m) == JVMClasspathPrepend:
                jb.cp = m.val + jb.cp
            elif type(m) == JSArg:
                logging.warning("JSArg not respected by JavaBenchmark")
        return jb

    def get_full_args(self, runtime: Runtime) -> List[Union[str, Path]]:
        cmd = super().get_full_args(runtime)
        cmd.append(runtime.get_executable())
        cmd.extend(self.jvm_args)
        cmd.extend(self.get_classpath_args())
        cmd.extend(self.program_args)
        return cmd


class JavaScriptBenchmark(Benchmark):
    def __init__(
        self, js_args: List[str], program: str, program_args: List[str], **kwargs
    ):
        super().__init__(**kwargs)
        self.js_args = js_args
        self.program = program
        self.program_args = program_args

    def __str__(self) -> str:
        return self.to_string(DummyRuntime("js"))

    def attach_modifiers(self, modifiers: Sequence[Modifier]) -> "JavaScriptBenchmark":
        jb = super().attach_modifiers(modifiers)
        for m in modifiers:
            if not m.should_attach(self.suite_name, self.name):
                continue
            if type(m) == ProgramArg:
                jb.program_args.extend(m.val)
            elif type(m) == JVMArg:
                logging.warning("JVMArg not respected by JavaScriptBenchmark")
            elif isinstance(m, JVMClasspathAppend) or type(m) == JVMClasspathPrepend:
                logging.warning("JVMClasspath not respected by JavaScriptBenchmark")
            elif type(m) == JSArg:
                jb.js_args.extend(m.val)
        return jb

    def get_full_args(self, runtime: Runtime) -> List[Union[str, Path]]:
        cmd = super().get_full_args(runtime)
        cmd.append(runtime.get_executable())
        cmd.extend(self.js_args)
        cmd.append(self.program)
        if isinstance(runtime, D8):
            cmd.append("--")
        elif isinstance(runtime, JavaScriptCore):
            cmd.append("--")
        elif isinstance(runtime, SpiderMonkey):
            pass
        else:
            raise TypeError(
                "{} is of type {}, and not a valid runtime for JavaScriptBenchmark".format(
                    runtime, type(runtime)
                )
            )
        cmd.extend(self.program_args)
        return cmd


class JuliaBenchmark(Benchmark):
    def __init__(
        self, julia_args: List[str], suite_path: Path, program_args: List[str], **kwargs
    ):
        super().__init__(**kwargs)
        self.julia_args = julia_args
        self.suite_path = suite_path
        self.program_args = program_args

    def __str__(self) -> str:
        return self.to_string(DummyRuntime("julia"))

    def attach_modifiers(self, modifiers: Sequence[Modifier]) -> "JuliaBenchmark":
        jb = super().attach_modifiers(modifiers)
        for m in modifiers:
            if type(m) == JuliaArg:
                jb.julia_args.extend(m.val)
            elif type(m) == ProgramArg:
                jb.program_args.extend(m.val)
        return jb

    def get_full_args(self, runtime: Runtime) -> List[Union[str, Path]]:
        cmd = super().get_full_args(runtime)
        cmd.append(runtime.get_executable())
        cmd.extend(self.julia_args)
        cmd.append("--project={}".format(self.suite_path))
        cmd.append(str(self.suite_path / "run_benchmarks.jl"))
        cmd.extend(self.name.split("/"))
        cmd.extend(["-n", "1"])  # one run
        cmd.extend(self.program_args)
        return cmd



from typing import Optional, TYPE_CHECKING
import zulip
from running.plugin.runbms import RunbmsPlugin
from running.util import Moma, register, MomaReservationStatus, config_index_to_chr
import logging
import copy
from running.suite import is_dry_run
from running.command.runbms import hfac_str
from datetime import datetime, timedelta

if TYPE_CHECKING:
    from running.benchmark import Benchmark

RESERVATION_WARNING_HOURS = 12
RESERVATION_WARNING_THRESHOLD = timedelta(seconds=RESERVATION_WARNING_HOURS * 60 * 60)


@register(RunbmsPlugin)
class Zulip(RunbmsPlugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.config_file = kwargs.get("config_file", "~/.zuliprc")
        self.client = zulip.Client(config_file=self.config_file)
        self.request = kwargs.get("request", {})
        if type(self.request) is not dict:
            raise TypeError("request of Zulip must be a dictionary")
        if self.request.get("type") not in ["private", "stream"]:
            raise ValueError("Request type must be either private or stream")
        if self.request.get("type") == "stream" and "topic" not in self.request:
            raise KeyError("Stream messages must have a topic")
        if "to" not in self.request:
            raise KeyError("Request must have a to field")
        self.nop = is_dry_run()
        self.moma = Moma()
        self.last_message_id = None
        self.last_message_content = None

    def send_message(self, content):
        message_data = copy.deepcopy(self.request)
        message_data["content"] = "{}\n{}{}\n".format(
            self.run_id, self.get_reservation_message(), content
        )
        try:
            result = self.client.send_message(message_data=message_data)
            if result["result"] != "success":
                logging.warning("Zulip send_message failed\n{}".format(result))
            else:
                self.last_message_id = result["id"]
                self.last_message_content = message_data["content"]
        except:
            logging.exception("Unhandled Zulip send_message exception")

    def modify_message(self, content):
        request = {
            "message_id": self.last_message_id,
            "content": content,
        }
        try:
            result = self.client.update_message(request)
            if result["result"] != "success":
                logging.warning("Zulip update_message failed\n{}".format(result))
            else:
                self.last_message_content = content
        except:
            logging.exception("Unhandled Zulip update_message exception")

    def __str__(self) -> str:
        return "Zulip {}".format(self.name)

    def start_hfac(self, hfac: Optional[float]):
        if self.nop:
            return
        self.send_message(
            "hfac {} started".format(hfac_str(hfac) if hfac is not None else "None")
        )

    def end_hfac(self, hfac: Optional[float]):
        if self.nop:
            return
        self.send_message(
            "hfac {} ended".format(hfac_str(hfac) if hfac is not None else "None")
        )

    def start_benchmark(
        self, _hfac: Optional[float], _size: Optional[int], bm: "Benchmark"
    ):
        if self.nop:
            return
        self.send_message("benchmark {} started".format(bm.name))

    def end_benchmark(
        self, _hfac: Optional[float], _size: Optional[int], bm: "Benchmark"
    ):
        if self.nop:
            return
        self.send_message("benchmark {} ended".format(bm.name))

    def start_invocation(
        self,
        _hfac: Optional[float],
        _size: Optional[int],
        _bm: "Benchmark",
        invocation: int,
    ):
        if self.nop:
            return
        if self.last_message_id and self.last_message_content:
            self.modify_message(self.last_message_content + str(invocation))

    def end_invocation(
        self,
        _hfac: Optional[float],
        _size: Optional[int],
        _bm: "Benchmark",
        _invocation: int,
    ):
        if self.nop:
            return

    def start_config(
        self,
        _hfac: Optional[float],
        _size: Optional[int],
        _bm: "Benchmark",
        _invocation: int,
        _config: str,
        _config_index: int,
    ):
        if self.nop:
            return

    def end_config(
        self,
        _hfac: Optional[float],
        _size: Optional[int],
        _bm: "Benchmark",
        _invocation: int,
        _config: str,
        config_index: int,
        passed: bool,
    ):
        if self.nop:
            return
        if self.last_message_id and self.last_message_content:
            if passed:
                self.modify_message(
                    self.last_message_content + config_index_to_chr(config_index)
                )
            else:
                self.modify_message(self.last_message_content + ".")

    def get_reservation_message(self) -> str:
        reservation = self.moma.get_reservation()
        if reservation is None:
            return ""
        if reservation.status is MomaReservationStatus.NOT_MOMA:
            return "# ** Warning: not running on a moma machine. **\n"
        elif reservation.status is MomaReservationStatus.NOT_RESERVED:
            return "# ** Warning: machine not reserved. **\n"
        elif reservation.status is MomaReservationStatus.RESERVED_BY_OTHERS:
            return "# ** Warning: machine reserved by {}, ends at {}. **\n".format(
                reservation.user, reservation.end
            )
        elif reservation.status is MomaReservationStatus.RESERVED_BY_ME:
            assert reservation.end is not None
            delta = reservation.end - datetime.now()
            if delta < RESERVATION_WARNING_THRESHOLD:
                return "# ** Warning: less than {} hours of reservation left. Current reservation ends at {}. **\n".format(
                    RESERVATION_WARNING_HOURS, reservation.end
                )
            else:
                return ""
        else:
            raise ValueError("Unhandled reservation status value")



from typing import Any, Dict, Optional, DefaultDict
from running.config import Configuration
from pathlib import Path
from running.runtime import NativeExecutable, Runtime
from running.benchmark import Benchmark, SubprocessrExit
from running.suite import BenchmarkSuite
from running.util import parse_config_str, config_str_encode
import logging
import tempfile
import yaml
from running.suite import is_dry_run
from collections import defaultdict
from enum import Enum
import os

configuration: Configuration


def setup_parser(subparsers):
    f = subparsers.add_parser("minheap")
    f.set_defaults(which="minheap")
    f.add_argument("CONFIG", type=Path)
    f.add_argument("RESULT", type=Path)
    f.add_argument("-a", "--attempts", type=int)


class ContinueSearch(Enum):
    Abort = 1
    HeapTooBig = 2
    HeapTooSmall = 3


def run_bm_with_retry(
    suite: BenchmarkSuite,
    runtime: Runtime,
    bm_with_heapsize: Benchmark,
    minheap_dir: Path,
    attempts: int,
) -> ContinueSearch:
    def log(s):
        return print(s, end="", flush=True)

    log(" ")
    for _ in range(attempts):
        output, _companion_output, subprocess_exit = bm_with_heapsize.run(
            runtime, cwd=minheap_dir
        )
        if runtime.is_oom(output):
            # if OOM is detected, we exit the loop regardless the exit statussour
            log("x ")
            return ContinueSearch.HeapTooSmall
        if subprocess_exit is SubprocessrExit.Normal:
            if suite.is_passed(output):
                log("o ")
                return ContinueSearch.HeapTooBig
        elif subprocess_exit is SubprocessrExit.Timeout:
            # A timeout is likely due to heap being too small and many GCs scheduled back to back
            log("t ")
            return ContinueSearch.HeapTooSmall
        # If not the above scenario, we treat this invocation as a crash or some kind of erroneous state
        log(".")
        continue
    # No successful invocation in the above attempts, but none OOMed either
    # Probably too many crashes, abort the binary search for this benchmark
    log(" ")
    return ContinueSearch.Abort


def minheap_one_bm(
    suite: BenchmarkSuite,
    runtime: Runtime,
    bm: Benchmark,
    heap: int,
    minheap_dir: Path,
    attempts: int,
) -> float:
    lo = 2
    hi = heap
    mid = (lo + hi) // 2
    minh = float("inf")
    while hi - lo > 1:
        heapsize = runtime.get_heapsize_modifiers(mid)
        size_str = "{}M".format(mid)
        print(size_str, end="", flush=True)
        bm_with_heapsize = bm.attach_modifiers(heapsize)
        result = run_bm_with_retry(
            suite, runtime, bm_with_heapsize, minheap_dir, attempts
        )
        if result is ContinueSearch.Abort:
            return float("inf")
        elif result is ContinueSearch.HeapTooBig:
            minh = mid
            hi = mid
            mid = (lo + hi) // 2
        elif result is ContinueSearch.HeapTooSmall:
            lo = mid
            mid = (lo + hi) // 2
        else:
            raise ValueError("Unhandled ContinueSearch variant")
    return minh


def run_with_persistence(
    result: Dict[str, Any],
    minheap_dir: Path,
    result_file: Optional[Path],
    attempts: int,
):
    suites = configuration.get("suites")
    maxheap = configuration.get("maxheap")
    for c in configuration.get("configs"):
        c_encoded = config_str_encode(c)
        if c_encoded not in result:
            result[c_encoded] = {}
        runtime, mods = parse_config_str(configuration, c)
        print("{} ".format(c_encoded))
        if isinstance(runtime, NativeExecutable):
            logging.warning("Minheap measurement not supported for NativeExecutable")
            continue
        for suite_name, bms in configuration.get("benchmarks").items():
            if suite_name not in result[c_encoded]:
                result[c_encoded][suite_name] = {}
            suite = suites[suite_name]
            for b in bms:
                # skip a benchmark if we have measured it
                if b.name in result[c_encoded][suite_name]:
                    continue
                print("\t {}-{} ".format(b.suite_name, b.name), end="")
                mod_b = b.attach_modifiers(mods)
                mod_b = mod_b.attach_modifiers(
                    b.get_runtime_specific_modifiers(runtime)
                )
                minheap = minheap_one_bm(
                    suite, runtime, mod_b, maxheap, minheap_dir, attempts
                )
                print("minheap {}".format(minheap))
                result[c_encoded][suite_name][b.name] = minheap
                if result_file:
                    with result_file.open("w") as fd:
                        yaml.dump(result, fd)


def print_best(result: Dict[str, Dict[str, Dict[str, float]]]):
    minheap: DefaultDict[str, DefaultDict[str, float]]
    minheap = defaultdict(lambda: defaultdict(lambda: float("inf")))
    minheap_config: DefaultDict[str, DefaultDict[str, str]]
    minheap_config = defaultdict(lambda: defaultdict(lambda: "ALL_FAILED"))
    for config, suites in result.items():
        for suite, benchmark_heap_sizes in suites.items():
            for benchmark, heap_size in benchmark_heap_sizes.items():
                if heap_size < minheap[suite][benchmark]:
                    minheap[suite][benchmark] = heap_size
                    minheap_config[suite][benchmark] = config

    config_best_count: DefaultDict[str, int]
    config_best_count = defaultdict(int)
    for suite, benchmark_configs in minheap_config.items():
        for benchmark, best_config in benchmark_configs.items():
            config_best_count[best_config] += 1

    if config_best_count.items():
        config, count = max(config_best_count.items(), key=lambda x: x[1])
        print(
            "{} obtained the most number of smallest minheap sizes: {}".format(
                config, count
            )
        )
        print("Minheap configuration to be copied to runbms config files")
        print(yaml.dump(result[config]))


def run(args):
    if args.get("which") != "minheap":
        return False
    global configuration
    configuration = Configuration.from_file(Path(os.getcwd()), args.get("CONFIG"))
    configuration.resolve_class()
    result_file = args.get("RESULT")
    if result_file.exists():
        with result_file.open() as fd:
            result = yaml.safe_load(fd)
            if result is None:
                result = {}
    else:
        result = {}
    attempts = configuration.get("attempts")
    if args.get("attempts"):
        attempts = args.get("attempts")
    with tempfile.TemporaryDirectory(prefix="minheap-") as minheap_dir:
        logging.info("Temporary directory: {}".format(minheap_dir))
        if is_dry_run():
            run_with_persistence(result, Path(minheap_dir), None, attempts)
        else:
            run_with_persistence(result, Path(minheap_dir), result_file, attempts)
    print_best(result)
    return True



from pathlib import Path
from running.config import Configuration
from running.util import parse_config_str, smart_quote, split_quoted


def test_split_quoted():
    assert split_quoted('123 "foo bar"') == ["123", "foo bar"]


def test_smart_quote():
    assert smart_quote(Path("/bin") / "123 456") == '"/bin/123 456"'


def test_issue104():
    c = Configuration(
        {
            "suites": {
                "dacapo2006": {
                    "type": "DaCapo",
                    "release": "2006",
                    "path": "/usr/share/benchmarks/dacapo/dacapo-2006-10-MR2.jar",
                    "timing_iteration": 3,
                }
            },
            "benchmarks": {"dacapo2006": ["fop"]},
            "runtimes": {
                "jdk8": {
                    "type": "OpenJDK",
                    "release": 8,
                    "home": "/usr/lib/jvm/temurin-8-jdk-amd64",
                }
            },
            "modifiers": {},
        }
    )
    c.resolve_class()
    _, modifiers = parse_config_str(c, "jdk8|")
    assert len(modifiers) == 0



from copy import deepcopy
from pathlib import Path
import gzip
import enum
from typing import Any, Callable, Dict, List
import functools
import re
from running.config import Configuration
import os

MMTk_HEADER = (
    "============================ MMTk Statistics Totals ============================"
)
MMTk_FOOTER = (
    "------------------------------ End MMTk Statistics -----------------------------"
)


class EditingMode(enum.Enum):
    NotEditing = 1
    Names = 2
    Values = 3
    TotalTime = 4


def setup_parser(subparsers):
    f = subparsers.add_parser("preproc")
    f.set_defaults(which="preproc")
    f.add_argument("CONFIG", type=Path)
    f.add_argument("SOURCE", type=Path)
    f.add_argument("TARGET", type=Path)


def filter_stats(predicate: Callable[[str], bool]):
    def inner(stats: Dict[str, float]):
        return {k: v for (k, v) in stats.items() if predicate(k)}

    return inner


def reduce_stats(pattern: str, new_column: str, func):
    compiled = re.compile(pattern)

    def inner(stats: Dict[str, float]):
        to_reduce = [v for (k, v) in stats.items() if compiled.match(k)]
        if not to_reduce:
            return stats
        new_stats = deepcopy(stats)
        new_stats[new_column] = functools.reduce(func, to_reduce)
        return new_stats

    return inner


def sum_work_perf_event(event_name):
    pattern = "work\\.\\w+\\.{}\\.total".format(event_name)
    new_column = "work.{}.total".format(event_name)
    return reduce_stats(pattern, new_column, lambda x, y: x + y)


def ratio_work_perf_event(event_name: str):
    pattern = "work\\.\\w+\\.{}\\.total".format(event_name)
    aggregated_column = "work.{}.total".format(event_name)
    compiled = re.compile(pattern)

    def inner(stats: Dict[str, float]):
        new_stats = deepcopy(stats)
        for k, v in stats.items():
            if compiled.match(k):
                new_column = k.replace(".total", ".ratio")
                new_stats[new_column] = v / stats[aggregated_column]
        return new_stats

    return inner


def ratio_event(event_name: str):
    def inner(stats: Dict[str, float]):
        new_stats = deepcopy(stats)
        stw_key = "{}.stw".format(event_name)
        other_key = "{}.other".format(event_name)
        if stw_key in stats and other_key in stats:
            gc = stats[stw_key]
            mu = stats[other_key]
            total = gc + mu
            new_stats["{}.ratio".format(stw_key)] = gc / total
            new_stats["{}.ratio".format(other_key)] = mu / total
        return new_stats

    return inner


def calc_ipc(stats: Dict[str, float]):
    new_stats = deepcopy(stats)
    for phase in ["mu", "gc"]:
        inst = stats.get("PERF_COUNT_HW_INSTRUCTIONS.{}".format(phase))
        cycles = stats.get("PERF_COUNT_HW_CPU_CYCLES.{}".format(phase))
        if inst is not None and cycles is not None:
            if cycles == 0:
                assert inst == 0
                continue
            new_stats["INSTRUCTIONS_PER_CYCLE.{}".format(phase)] = inst / cycles
    return new_stats


def calc_work_ipc(stats: Dict[str, float]):
    pattern = "work\\.\\w+\\.PERF_COUNT_HW_INSTRUCTIONS\\.total"
    compiled = re.compile(pattern)
    new_stats = deepcopy(stats)
    for k, v in stats.items():
        if compiled.match(k):
            cycles = k.replace("PERF_COUNT_HW_INSTRUCTIONS", "PERF_COUNT_HW_CPU_CYCLES")
            ipc = k.replace(
                "PERF_COUNT_HW_INSTRUCTIONS.total", "INSTRUCTIONS_PER_CYCLE.ratio"
            )
            new_stats[ipc] = stats[k] / stats[cycles]
    return new_stats


def stat_sort_helper(key: str, value: float):
    if len(key.split(".")) > 1:
        return key.split(".")[-2], -value
    else:
        return key, -value


def process_lines(configuration: Configuration, lines: List[str]):
    new_lines = []
    editing = EditingMode.NotEditing
    names = []
    funcs: List[Any]
    funcs = []
    if configuration.get("preprocessing") is None:
        funcs = []
    else:
        for f in configuration.get("preprocessing"):
            if f["name"] == "sum_work_perf_event":
                for v in f["val"].split(","):
                    funcs.append(sum_work_perf_event(v))
            elif f["name"] == "ratio_work_perf_event":
                for v in f["val"].split(","):
                    funcs.append(ratio_work_perf_event(v))
            elif f["name"] == "calc_work_ipc":
                funcs.append(calc_work_ipc)
            elif f["name"] == "ratio_event":
                for v in f["val"].split(","):
                    funcs.append(ratio_event(v))
            elif f["name"] == "filter_stats":
                patterns_to_keep = f["val"].split(",")
                funcs.append(
                    filter_stats(lambda n: any([p in n for p in patterns_to_keep]))
                )
            elif f["name"] == "calc_ipc":
                funcs.append(calc_ipc)
            else:
                raise ValueError("Not supported preprocessing functionality")
    for line in lines:
        if line.strip() == MMTk_HEADER:
            new_lines.append(line)
            editing = EditingMode.Names
            continue
        elif line.strip() == MMTk_FOOTER:
            assert editing == EditingMode.TotalTime
            new_lines.append(line)
            editing = EditingMode.NotEditing
            continue
        if editing == EditingMode.Names:
            names = line.strip().split("\t")
            editing = EditingMode.Values
        elif editing == EditingMode.Values:
            values = map(float, line.strip().split("\t"))
            stats = dict(zip(names, values))
            new_stats = functools.reduce(lambda accum, val: val(accum), funcs, stats)
            if len(new_stats):
                new_stat_list = list(new_stats.items())
                new_stat_list.sort(key=lambda x: stat_sort_helper(x[0], x[1]))
                new_names, new_values = list(zip(*new_stat_list))
                new_lines.append("{}\n".format("\t".join(new_names)))
                new_lines.append("{}\n".format("\t".join(map(str, new_values))))
            else:
                new_lines.append("empty_after_preprocessing\n")
                new_lines.append("0\n")
            editing = EditingMode.TotalTime
        elif editing == EditingMode.TotalTime:
            new_lines.append(line)
        else:
            new_lines.append(line)

    return new_lines


def process_one_file(configuration: Configuration, original: Path, targetfile: Path):
    # XXX DO NOT COPY the content of the log file
    # Tab might not be preserved (especially around line breaks)
    # https://unix.stackexchange.com/questions/324676/output-tab-character-on-terminal-window
    with gzip.open(original, "rt") as old:
        with gzip.open(targetfile, "wt") as new:
            new.writelines(process_lines(configuration, old.readlines()))


def process(configuration: Configuration, source: Path, target: Path):
    for file in source.glob("*.log.gz"):
        process_one_file(configuration, file, target / file.name)


def run(args):
    if args.get("which") != "preproc":
        return False
    configuration = Configuration.from_file(Path(os.getcwd()), args.get("CONFIG"))
    source = args.get("SOURCE")
    target = args.get("TARGET")
    target.mkdir(parents=True, exist_ok=True)
    process(configuration, source, target)
    return True



from pathlib import Path
from typing import Any, Dict, Optional, Union, List, Sequence
from running.benchmark import (
    JavaBenchmark,
    BinaryBenchmark,
    Benchmark,
    JavaScriptBenchmark,
    JuliaBenchmark,
)
from running.runtime import OpenJDK, Runtime
from running.modifier import JVMArg, Modifier
import logging
from running.util import register, split_quoted
import os.path

__DRY_RUN = False
__DEFAULT_MINHEAP = 4096


def is_dry_run():
    global __DRY_RUN
    return __DRY_RUN


def set_dry_run(val: bool):
    global __DRY_RUN
    __DRY_RUN = val


def parse_timing_iteration(t: Optional[str], suite_name: str) -> Union[str, int]:
    if not t:
        raise KeyError(
            "You need to specify the timing_iteration for a {} suite".format(suite_name)
        )
    assert t is not None
    try:
        t_parsed = int(t)
        return t_parsed
    except ValueError:
        return t


class BenchmarkSuite(object):
    CLS_MAPPING: Dict[str, Any]
    CLS_MAPPING = {}

    def __init__(self, name: str, **kwargs):
        self.name = name

    def __str__(self) -> str:
        return "Benchmark Suite {}".format(self.name)

    @staticmethod
    def from_config(name: str, config: Dict[str, str]) -> Any:
        return BenchmarkSuite.CLS_MAPPING[config["type"]](name=name, **config)

    def get_benchmark(self, _bm_spec: Union[str, Dict[str, Any]]) -> Any:
        raise NotImplementedError()

    def get_minheap(self, _bm: Benchmark) -> int:
        raise NotImplementedError

    def is_passed(self, _output: bytes) -> bool:
        raise NotImplementedError


@register(BenchmarkSuite)
class BinaryBenchmarkSuite(BenchmarkSuite):
    def __init__(self, programs: Dict[str, Dict[str, str]], **kwargs):
        super().__init__(**kwargs)
        self.programs: Dict[str, Dict[str, Any]]
        self.programs = {
            k: {"path": Path(v["path"]), "args": split_quoted(v["args"])}
            for k, v in programs.items()
        }
        self.timeout = kwargs.get("timeout")

    def get_benchmark(self, bm_spec: Union[str, Dict[str, Any]]) -> "BinaryBenchmark":
        assert type(bm_spec) is str
        bm_name = bm_spec
        return BinaryBenchmark(
            self.programs[bm_name]["path"],
            self.programs[bm_name]["args"],
            suite_name=self.name,
            name=bm_name,
            timeout=self.timeout,
        )

    def get_minheap(self, bm: Benchmark) -> int:
        logging.warning("minheap is not respected for BinaryBenchmarkSuite")
        assert isinstance(bm, BinaryBenchmark)
        return 0

    def is_passed(self, _output: bytes) -> bool:
        # FIXME no generic way to know
        return True


class JavaBenchmarkSuite(BenchmarkSuite):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def get_minheap(self, _bm: Benchmark) -> int:
        raise NotImplementedError()


@register(BenchmarkSuite)
class DaCapo(JavaBenchmarkSuite):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.release: str
        self.release = kwargs["release"]
        if self.release not in ["2006", "9.12", "evaluation", "23.11"]:
            raise ValueError("DaCapo release {} not recognized".format(self.release))
        self.path: Path
        self.path = Path(os.path.expandvars(kwargs["path"]))
        if not self.path.exists():
            logging.warning("DaCapo jar {} not found".format(self.path))
        self.minheap: Optional[str]
        self.minheap = kwargs.get("minheap")
        self.minheap_values: Dict[str, Dict[str, int]]
        self.minheap_values = kwargs.get("minheap_values", {})
        if not isinstance(self.minheap_values, dict):
            raise TypeError(
                "The minheap_values of {} should be a dictionary".format(self.name)
            )
        if self.minheap:
            if not isinstance(self.minheap, str):
                raise TypeError(
                    "The minheap of {} should be a string that selects from a minheap_values".format(
                        self.name
                    )
                )
            if self.minheap not in self.minheap_values:
                raise KeyError(
                    "{} is not a valid entry of {}.minheap_values".format(
                        self.name, self.name
                    )
                )
        self.timing_iteration = parse_timing_iteration(
            kwargs.get("timing_iteration"), "DaCapo"
        )
        if (
            isinstance(self.timing_iteration, str)
            and self.timing_iteration != "converge"
        ):
            raise TypeError(
                "The timing iteration of the DaCapo benchmark suite `{}` is {}, which neither an integer nor 'converge'".format(
                    self.path, repr(self.timing_iteration)
                )
            )
        self.callback: Optional[str]
        self.callback = kwargs.get("callback")
        self.timeout: Optional[int]
        self.timeout = kwargs.get("timeout")
        self.wrapper: Optional[Union[Dict[str, str], str]]
        self.wrapper = kwargs.get("wrapper")
        self.companion: Optional[Union[Dict[str, str], str]]
        self.companion = kwargs.get("companion")
        # user overriding the default size for the entire suite
        self.size: Optional[str]
        self.size = kwargs.get("size")

    def __str__(self) -> str:
        return "{} DaCapo {} {}".format(super().__str__(), self.release, self.path)

    @staticmethod
    def parse_timing_iteration(v: Any):
        try:
            timing_iteration = int(v)
        except ValueError:
            if v != "converge":
                raise TypeError(
                    "The timing iteration {} is neither an integer nor 'converge'".format(
                        repr(v)
                    )
                )
            timing_iteration = v
        return timing_iteration

    def get_benchmark(self, bm_spec: Union[str, Dict[str, Any]]) -> "JavaBenchmark":
        timing_iteration = self.timing_iteration
        timeout = self.timeout
        size = self.size
        if type(bm_spec) is str:
            bm_name = bm_spec
            name = bm_spec
        else:
            assert type(bm_spec) is dict
            if "bm_name" not in bm_spec or "name" not in bm_spec:
                raise KeyError(
                    "When a dictionary is used to speicfy a benchmark, you need to provide both `name` and `bm_name`"
                )
            bm_name = bm_spec["bm_name"]
            name = bm_spec["name"]
            if "timing_iteration" in bm_spec:
                timing_iteration = DaCapo.parse_timing_iteration(
                    bm_spec["timing_iteration"]
                )
            # user overriding the size for that benchmark
            if "size" in bm_spec:
                size = bm_spec["size"]
            if "timeout" in bm_spec:
                timeout = bm_spec["timeout"]

        if self.callback:
            cp = [str(self.path)]
            program_args = ["Harness", "-c", self.callback]
        else:
            cp = []
            program_args = ["-jar", str(self.path)]
        # Timing iteration
        if type(timing_iteration) is int:
            program_args.extend(["-n", str(timing_iteration)])
        else:
            assert timing_iteration == "converge"
            if self.release == "2006":
                program_args.append("-converge")
            else:
                program_args.append("--converge")
        # Input size
        if size:
            program_args.extend(["-s", size])
        # Name of the benchmark
        program_args.append(bm_name)

        return JavaBenchmark(
            jvm_args=[],
            program_args=program_args,
            cp=cp,
            wrapper=self.get_wrapper(bm_name),
            companion=self.get_companion(bm_name),
            suite_name=self.name,
            name=name,
            timeout=timeout,
        )

    def get_minheap(self, bm: Benchmark) -> int:
        assert isinstance(bm, JavaBenchmark)
        name = bm.name
        if not self.minheap:
            logging.warning("No minheap_value of {} is selected".format(self))
            return __DEFAULT_MINHEAP
        minheap = self.minheap_values[self.minheap]
        if name not in minheap:
            logging.warning("Minheap for {} of {} not set".format(name, self))
            return __DEFAULT_MINHEAP
        return minheap[name]

    def is_passed(self, output: bytes) -> bool:
        return b"PASSED" in output

    def get_wrapper(self, bm_name: str) -> Optional[str]:
        if self.wrapper is None:
            return None
        elif type(self.wrapper) == str:
            return self.wrapper
        elif type(self.wrapper) == dict:
            return self.wrapper.get(bm_name)
        else:
            raise TypeError(
                "wrapper of {} must be either null, "
                "a string (the same wrapper for all benchmarks), "
                "or a dictionary (different wrappers for"
                "differerent benchmarks)".format(self.name)
            )

    def get_companion(self, bm_name: str) -> Optional[str]:
        if self.companion is None:
            return None
        elif type(self.companion) == str:
            return self.companion
        elif type(self.companion) == dict:
            return self.companion.get(bm_name)
        else:
            raise TypeError(
                "companion of {} must be either null, "
                "a string (the same companion for all benchmarks), "
                "or a dictionary (different companions for"
                "differerent benchmarks)".format(self.name)
            )


@register(BenchmarkSuite)
class SPECjbb2015(JavaBenchmarkSuite):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.release: str
        self.release = kwargs["release"]
        if self.release not in ["1.03"]:
            raise ValueError(
                "SPECjbb2015 release {} not recognized".format(self.release)
            )
        self.path: Path
        self.path = Path(os.path.expandvars(kwargs["path"])).resolve()
        self.propsfile = (self.path / ".." / "config" / "specjbb2015.props").resolve()
        if not self.path.exists():
            logging.info("SPECjbb2015 jar {} not found".format(self.path))

    def __str__(self) -> str:
        return "{} SPECjbb2015 {} {}".format(super().__str__(), self.release, self.path)

    def get_benchmark(self, bm_spec: Union[str, Dict[str, Any]]) -> "JavaBenchmark":
        assert type(bm_spec) is str
        if bm_spec != "composite":
            raise ValueError("Only composite mode is supported for now")

        program_args = [
            "-jar",
            str(self.path),
            "-p",
            str(self.propsfile),
            "-m",
            "COMPOSITE",
            "-skipReport",
        ]
        return JavaBenchmark(
            jvm_args=[],
            program_args=program_args,
            cp=[],
            suite_name=self.name,
            name="composite",
        )

    def get_minheap(self, _bm: Benchmark) -> int:
        return 2048  # SPEC recommends running with minimum 2GB of heap

    def is_passed(self, output: bytes) -> bool:
        # FIXME
        return True


@register(BenchmarkSuite)
class Octane(BenchmarkSuite):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.path: Path
        self.path = Path(os.path.expandvars(kwargs["path"])).resolve()
        if not self.path.exists():
            logging.info("Octane folder {} not found".format(self.path))
        self.wrapper: Path
        self.wrapper = Path(kwargs["wrapper"]).resolve()
        if not self.wrapper.exists():
            logging.info("Octane folder {} not found".format(self.wrapper))
        timing_iteration = parse_timing_iteration(
            kwargs.get("timing_iteration"), "Octane"
        )
        self.timing_iteration: int
        if isinstance(timing_iteration, str):
            raise TypeError("timing_iteration for Octane has to be an integer")
        else:
            self.timing_iteration = timing_iteration
        self.minheap: Optional[str]
        self.minheap = kwargs.get("minheap")
        self.minheap_values: Dict[str, Dict[str, int]]
        self.minheap_values = kwargs.get("minheap_values", {})
        if not isinstance(self.minheap_values, dict):
            raise TypeError(
                "The minheap_values of {} should be a dictionary".format(self.name)
            )
        if self.minheap:
            if not isinstance(self.minheap, str):
                raise TypeError(
                    "The minheap of {} should be a string that selects from a minheap_values".format(
                        self.name
                    )
                )
            if self.minheap not in self.minheap_values:
                raise KeyError(
                    "{} is not a valid entry of {}.minheap_values".format(
                        self.name, self.name
                    )
                )
        self.timeout: Optional[int]
        self.timeout = kwargs.get("timeout")

    def __str__(self) -> str:
        return "{} Octane {}".format(super().__str__(), self.path)

    def get_benchmark(
        self, bm_spec: Union[str, Dict[str, Any]]
    ) -> "JavaScriptBenchmark":
        assert type(bm_spec) is str

        program_args = [str(self.path), bm_spec, str(self.timing_iteration)]
        return JavaScriptBenchmark(
            js_args=[],
            program=str(self.wrapper),
            program_args=program_args,
            suite_name=self.name,
            name=bm_spec,
            timeout=self.timeout,
        )

    def get_minheap(self, bm: Benchmark) -> int:
        assert isinstance(bm, JavaScriptBenchmark)
        name = bm.name
        if not self.minheap:
            logging.warning("No minheap_value of {} is selected".format(self))
            return __DEFAULT_MINHEAP
        minheap = self.minheap_values[self.minheap]
        if name not in minheap:
            logging.warning("Minheap for {} of {} not set".format(name, self))
            return __DEFAULT_MINHEAP
        return minheap[name]

    def is_passed(self, output: bytes) -> bool:
        return b"PASSED" in output


@register(BenchmarkSuite)
class SPECjvm98(JavaBenchmarkSuite):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.release: str
        self.release = kwargs["release"]
        if self.release not in ["1.03_05"]:
            raise ValueError("SPECjvm98 release {} not recognized".format(self.release))
        self.path: Path
        self.path = Path(os.path.expandvars(kwargs["path"])).resolve()

        if not self.path.exists():
            logging.info("SPECjvm98 {} not found".format(self.path))
        if not (self.path / "SpecApplication.class").exists():
            logging.info(
                "SpecApplication.class not found under SPECjvm98 {}".format(self.path)
            )
        timing_iteration = parse_timing_iteration(
            kwargs.get("timing_iteration"), "SPECjvm98"
        )
        self.timing_iteration: int
        if isinstance(timing_iteration, str):
            raise TypeError("timing_iteration for SPECjvm98 has to be an integer")
        else:
            self.timing_iteration = timing_iteration

    def __str__(self) -> str:
        return "{} SPECjvm98 {} {}".format(super().__str__(), self.release, self.path)

    def get_benchmark(self, bm_spec: Union[str, Dict[str, Any]]) -> "JavaBenchmark":
        assert type(bm_spec) is str
        program_args = [
            "SpecApplication",
            "-i{}".format(self.timing_iteration),
            bm_spec,
        ]
        return JavaBenchmark(
            jvm_args=[],
            program_args=program_args,
            cp=[str(self.path)],
            suite_name=self.name,
            name=bm_spec,
            override_cwd=self.path,
        )

    def get_minheap(self, _bm: Benchmark) -> int:
        # FIXME allow user to measure and specify minimum heap sizes
        return 32  # SPEC recommends running with minimum 32MB of heap

    def is_passed(self, output: bytes) -> bool:
        # FIXME
        return b"**NOT VALID**" not in output


@register(BenchmarkSuite)
class JuliaGCBenchmarks(BenchmarkSuite):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.path: Path
        self.path = Path(os.path.expandvars(kwargs["path"]))
        if not self.path.exists():
            logging.warning("JuliaGCBenchmarks does not exist at {}".format(self.path))
        self.minheap: Optional[str]
        self.minheap = kwargs.get("minheap")
        self.minheap_values: Dict[str, Dict[str, int]]
        self.minheap_values = kwargs.get("minheap_values", {})
        if not isinstance(self.minheap_values, dict):
            raise TypeError(
                "The minheap_values of {} should be a dictionary".format(self.name)
            )
        if self.minheap:
            if not isinstance(self.minheap, str):
                raise TypeError(
                    "The minheap of {} should be a string that selects from a minheap_values".format(
                        self.name
                    )
                )
            if self.minheap not in self.minheap_values:
                raise KeyError(
                    "{} is not a valid entry of {}.minheap_values".format(
                        self.name, self.name
                    )
                )

    def __str__(self) -> str:
        return "{} JuliaGCBenchmarks {}".format(super().__str__(), self.path)

    def get_minheap(self, bm: Benchmark) -> int:
        name = bm.name
        if not self.minheap:
            logging.warning("No minheap_value of {} is selected".format(self))
            return __DEFAULT_MINHEAP
        minheap = self.minheap_values[self.minheap]
        if name not in minheap:
            logging.warning("Minheap for {} of {} not set".format(name, self))
            return __DEFAULT_MINHEAP
        return minheap[name]

    def get_benchmark(self, bm_spec: Union[str, Dict[str, Any]]) -> "JuliaBenchmark":
        assert type(bm_spec) is str
        return JuliaBenchmark(
            julia_args=[],
            suite_name=self.name,
            name=bm_spec,
            suite_path=self.path,
            program_args=[],
        )

    def is_passed(self, output: bytes) -> bool:
        # FIXME
        return True



from running.command.runbms import spread
import pytest


def test_spread_0():
    spread_factor = 0
    N = 8
    for i in range(0, N + 1):
        assert spread(spread_factor, N, i) == i


def test_spread_1():
    spread_factor = 1
    N = 8
    for i in range(1, N + 1):
        left = pytest.approx(
            spread(spread_factor, N, i) - spread(spread_factor, N, i - 1)
        )
        right = pytest.approx(1 + (i - 1) / 7)
        assert left == right



from typing import Any, List, TYPE_CHECKING, Optional, Tuple, Set

if TYPE_CHECKING:
    from running.config import Configuration
    from running.modifier import Modifier
    from running.runtime import Runtime
import shlex
import socket
import urllib.request
import enum
import getpass
from datetime import datetime
import subprocess
import time


def system(cmd, check=True) -> str:
    return subprocess.run(
        cmd, check=check, stdout=subprocess.PIPE, shell=True
    ).stdout.decode("utf-8")


def register(parent_class):
    def inner(cls):
        parent_class.CLS_MAPPING[cls.__name__] = cls
        return cls

    return inner


def config_index_to_chr(i: int) -> str:
    if i < 0 or i >= 52:
        raise ValueError("Cannot convert {} into a character".format(i))
    elif i < 26:
        return chr(ord("a") + i)
    else:
        return chr(ord("A") + i - 26)


def parse_modifier_strs(
    configuration: "Configuration", mod_strs: List[str]
) -> List["Modifier"]:
    # Some modifiers could be a modifier set, and we need to flatten it
    from running.modifier import ModifierSet

    mods = []
    for m in mod_strs:
        m = m.strip()
        if not m:
            break
        mod_name = m.split("-")[0]
        mod_value_opts = m.split("-")[1:]
        mod = configuration.get("modifiers").get(mod_name)
        if mod is None:
            raise KeyError("Modifier '{}' not defined".format(mod_name))
        mod = mod.apply_value_opts(mod_value_opts)
        if isinstance(mod, ModifierSet):
            for m_inner in mod.flatten(configuration):
                mods.append(m_inner)
        else:
            mods.append(mod)
    return mods


def parse_config_str(
    configuration: "Configuration", c: str
) -> Tuple["Runtime", List["Modifier"]]:
    runtime = configuration.get("runtimes")[c.split("|")[0].strip()]
    mods = parse_modifier_strs(configuration, c.split("|")[1:])
    return runtime, mods


def config_str_encode(c: str) -> str:
    return ".".join([x.strip() for x in c.split("|")])


def split_quoted(s: str) -> List[str]:
    return shlex.split(s)


def smart_quote(_s: Any) -> str:
    s = str(_s)
    if s == "":
        return '""'
    need_quote = False
    for c in s:
        if not (c.isalnum() or c in ".:/+=-_"):
            need_quote = True
            break
    if need_quote:
        return '"{}"'.format(s)
    else:
        return s


def get_logged_in_users() -> Set[str]:
    output = system("who")
    return set([l.split()[0] for l in output.splitlines()])


class MomaReservationStatus(enum.Enum):
    NOT_RESERVED = 1
    RESERVED_BY_OTHERS = 2
    RESERVED_BY_ME = 3
    NOT_MOMA = 4


class MomaReservaton(object):
    def __init__(
        self,
        status: MomaReservationStatus,
        user: Optional[str],
        start: Optional[datetime],
        end: Optional[datetime],
    ):
        self.status = status
        self.user = user
        self.start = start
        self.end = end


class Moma(object):
    def __init__(self, host: Optional[str] = None, frequency: int = 60):
        if host is None:
            self.host = system("hostname -s").strip()
            self.is_moma = system("hostname -d").strip() == "moma"
        else:
            self.host = host
            try:
                self.is_moma = socket.gethostbyname_ex(self.host)[0].endswith(".moma")
            except socket.gaierror:
                self.is_moma = False
        self.reserve_time_url = "http://10.0.0.1/reserve-time?host={}".format(self.host)
        self.frequency = frequency
        self.last_checked: Optional[float]
        self.last_checked = None
        self.reservation: Optional[MomaReservaton]
        self.reservation = None
        self.update_reservation()

    def update_reservation(self):
        now = time.time()
        if self.last_checked:
            if now - self.last_checked <= self.frequency:
                return
        if not self.is_moma:
            self.reservation = MomaReservaton(
                MomaReservationStatus.NOT_MOMA, None, None, None
            )
        else:
            with urllib.request.urlopen(self.reserve_time_url) as response:
                html = response.read()
                if not html:
                    self.reservation = MomaReservaton(
                        MomaReservationStatus.NOT_RESERVED, None, None, None
                    )
                else:
                    html = html.decode("utf-8")
                    user, start, end = html.split(",")
                    status = (
                        MomaReservationStatus.RESERVED_BY_ME
                        if user == getpass.getuser()
                        else MomaReservationStatus.RESERVED_BY_OTHERS
                    )
                    start = datetime.fromtimestamp(int(start))
                    end = datetime.fromtimestamp(int(end))
                    self.reservation = MomaReservaton(status, user, start, end)
        self.last_checked = now

    def get_reservation(self) -> Optional[MomaReservaton]:
        self.update_reservation()
        return self.reservation
